# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from collections import defaultdict
from typing import ClassVar
from typing import Sequence

from twisted.internet import defer
from twisted.python import log
from twisted.python.deprecate import deprecatedModuleAttribute
from twisted.python.versions import Version

from buildbot import config
from buildbot import util
from buildbot.changes import changes
from buildbot.changes.filter import ChangeFilter
from buildbot.schedulers import base
from buildbot.schedulers import dependent
from buildbot.util import NotABranch
from buildbot.util.codebase import AbsoluteSourceStampsMixin
from buildbot.warnings import warn_deprecated


class BaseBasicScheduler(base.ReconfigurableBaseScheduler):
    """
    @param onlyImportant: If True, only important changes will be added to the
                          buildset.
    @type onlyImportant: boolean

    """

    compare_attrs: ClassVar[Sequence[str]] = (
        'treeStableTimer',
        'change_filter',
        'fileIsImportant',
        'onlyImportant',
        'reason',
    )

    fileIsImportant = None
    reason = ''

    class NotSet:
        pass

    def __init__(self, name, **kwargs):
        super().__init__(name=name, **kwargs)
        # the IDelayedCall used to wake up when this scheduler's
        # treeStableTimer expires.
        self._stable_timers = defaultdict(lambda: None)
        self._stable_timers_lock = defer.DeferredLock()

    def checkConfig(  # type: ignore[override]
        self,
        shouldntBeSet=NotSet,
        treeStableTimer=None,
        builderNames=None,
        branch=NotABranch,
        branches=NotABranch,
        fileIsImportant=None,
        categories=None,
        reason="The %(classname)s scheduler named '%(name)s' triggered this build",
        change_filter=None,
        onlyImportant=False,
        **kwargs,
    ):
        if shouldntBeSet is not self.NotSet:
            config.error("pass arguments to schedulers using keyword arguments")
        if fileIsImportant and not callable(fileIsImportant):
            config.error("fileIsImportant must be a callable")

        super().checkConfig(builderNames=builderNames, **kwargs)

    @defer.inlineCallbacks
    def reconfigService(  # type: ignore[override]
        self,
        shouldntBeSet=NotSet,
        treeStableTimer=None,
        builderNames=None,
        branch=NotABranch,
        branches=NotABranch,
        fileIsImportant=None,
        categories=None,
        reason="The %(classname)s scheduler named '%(name)s' triggered this build",
        change_filter=None,
        onlyImportant=False,
        **kwargs,
    ):
        yield super().reconfigService(builderNames=builderNames, **kwargs)

        self.treeStableTimer = treeStableTimer
        self.fileIsImportant = fileIsImportant
        self.onlyImportant = onlyImportant
        self.change_filter = self.getChangeFilter(
            branch=branch, branches=branches, change_filter=change_filter, categories=categories
        )

        self.reason = util.bytes2unicode(
            reason % {'name': self.name, 'classname': self.__class__.__name__}
        )

        if self.active:
            yield self._stopConsumingChanges()
            yield self.startConsumingChanges(
                fileIsImportant=self.fileIsImportant,
                change_filter=self.change_filter,
                onlyImportant=self.onlyImportant,
            )

    def getChangeFilter(self, branch, branches, change_filter, categories):
        raise NotImplementedError

    @defer.inlineCallbacks
    def activate(self):
        yield super().activate()

        if not self.enabled:
            return

        yield self.startConsumingChanges(
            fileIsImportant=self.fileIsImportant,
            change_filter=self.change_filter,
            onlyImportant=self.onlyImportant,
        )

        # if we have a treeStableTimer, if there are classified changes
        # out there, start their timers again
        if self.treeStableTimer:
            yield self.scanExistingClassifiedChanges()

        # otherwise, we don't care about classified
        # changes, so get rid of any hanging around from previous
        # configurations
        else:
            yield self.master.db.schedulers.flushChangeClassifications(self.serviceid)

    @defer.inlineCallbacks
    def deactivate(self):
        # the base deactivate will unsubscribe from new changes
        yield super().deactivate()

        if not self.enabled:
            return

        @util.deferredLocked(self._stable_timers_lock)
        def cancel_timers():
            for timer in self._stable_timers.values():
                if timer:
                    timer.cancel()
            self._stable_timers.clear()

        yield cancel_timers()

    @util.deferredLocked('_stable_timers_lock')
    def gotChange(self, change, important):
        if not self.treeStableTimer:
            # if there's no treeStableTimer, we can completely ignore
            # unimportant changes
            if not important:
                return defer.succeed(None)
            # otherwise, we'll build it right away
            return self.addBuildsetForChanges(
                reason=self.reason, changeids=[change.number], priority=self.priority
            )

        timer_name = self.getTimerNameForChange(change)

        # if we have a treeStableTimer
        # - for an important change, start the timer
        # - for an unimportant change, reset the timer if it is running

        if important or self._stable_timers[timer_name]:
            if self._stable_timers[timer_name]:
                self._stable_timers[timer_name].cancel()

            def fire_timer():
                d = self.stableTimerFired(timer_name)
                d.addErrback(log.err, "while firing stable timer")

            self._stable_timers[timer_name] = self.master.reactor.callLater(
                self.treeStableTimer, fire_timer
            )

        # record the change's importance
        return self.master.db.schedulers.classifyChanges(self.serviceid, {change.number: important})

    @defer.inlineCallbacks
    def scanExistingClassifiedChanges(self):
        # call gotChange for each classified change.  This is called at startup
        # and is intended to re-start the treeStableTimer for any changes that
        # had not yet been built when the scheduler was stopped.

        # NOTE: this may double-call gotChange for changes that arrive just as
        # the scheduler starts up.  In practice, this doesn't hurt anything.
        classifications = yield self.master.db.schedulers.getChangeClassifications(self.serviceid)

        # call gotChange for each change, after first fetching it from the db
        for changeid, important in classifications.items():
            chdict = yield self.master.db.changes.getChange(changeid)

            if not chdict:
                continue

            change = yield changes.Change.fromChdict(self.master, chdict)
            yield self.gotChange(change, important)

    def getTimerNameForChange(self, change):
        raise NotImplementedError  # see subclasses

    def getChangeClassificationsForTimer(self, sched_id, timer_name):
        """similar to db.schedulers.getChangeClassifications, but given timer
        name"""
        raise NotImplementedError  # see subclasses

    @util.deferredLocked('_stable_timers_lock')
    @defer.inlineCallbacks
    def stableTimerFired(self, timer_name):
        # delete this now-fired timer, if the service has already been stopped
        # then just bail out
        if not self._stable_timers.pop(timer_name, None):
            return

        classifications = yield self.getChangeClassificationsForTimer(self.serviceid, timer_name)

        # just in case: databases do weird things sometimes!
        if not classifications:  # pragma: no cover
            return

        changeids = sorted(classifications.keys())
        yield self.addBuildsetForChanges(
            reason=self.reason, changeids=changeids, priority=self.priority
        )

        max_changeid = changeids[-1]  # (changeids are sorted)
        yield self.master.db.schedulers.flushChangeClassifications(
            self.serviceid, less_than=max_changeid + 1
        )


class SingleBranchScheduler(AbsoluteSourceStampsMixin, BaseBasicScheduler):
    createAbsoluteSourceStamps = None

    def checkConfig(self, createAbsoluteSourceStamps=False, **kwargs):
        super().checkConfig(**kwargs)

    @defer.inlineCallbacks
    def reconfigService(
        self,
        createAbsoluteSourceStamps=False,
        **kwargs,
    ):
        self.createAbsoluteSourceStamps = createAbsoluteSourceStamps
        yield super().reconfigService(**kwargs)

    @defer.inlineCallbacks
    def gotChange(self, change, important):
        if self.createAbsoluteSourceStamps:
            yield self.recordChange(change)

        yield super().gotChange(change, important)

    def getCodebaseDict(self, codebase):
        if self.createAbsoluteSourceStamps:
            return super().getCodebaseDict(codebase)
        return self.codebases[codebase]

    def getChangeFilter(self, branch, branches, change_filter, categories):
        if branch is NotABranch and not change_filter:
            config.error(
                "the 'branch' argument to SingleBranchScheduler is "
                + "mandatory unless change_filter is provided"
            )
        elif branches is not NotABranch:
            config.error("the 'branches' argument is not allowed for " + "SingleBranchScheduler")

        return ChangeFilter.fromSchedulerConstructorArgs(
            change_filter=change_filter, branch=branch, categories=categories
        )

    def getTimerNameForChange(self, change):
        return "only"  # this class only uses one timer

    def getChangeClassificationsForTimer(self, sched_id, timer_name):
        return self.master.db.schedulers.getChangeClassifications(sched_id)


class Scheduler(SingleBranchScheduler):
    "alias for SingleBranchScheduler"

    def __init__(self, *args, **kwargs):
        warn_deprecated(
            '4.3.0',
            (
                'the name \'Scheduler\' is deprecated; use '
                'buildbot.schedulers.basic.SingleBranchScheduler instead'
            ),
        )
        super().__init__(*args, **kwargs)


class AnyBranchScheduler(BaseBasicScheduler):
    def getChangeFilter(self, branch, branches, change_filter, categories):
        assert branch is NotABranch
        return ChangeFilter.fromSchedulerConstructorArgs(
            change_filter=change_filter, branch=branches, categories=categories
        )

    def getTimerNameForChange(self, change):
        # Py2.6+: could be a namedtuple
        return (change.codebase, change.project, change.repository, change.branch)

    def getChangeClassificationsForTimer(self, sched_id, timer_name):
        # set in getTimerNameForChange
        codebase, project, repository, branch = timer_name
        return self.master.db.schedulers.getChangeClassifications(
            sched_id, branch=branch, repository=repository, codebase=codebase, project=project
        )


# now at buildbot.schedulers.dependent, but keep the old name alive
Dependent = dependent.Dependent

deprecatedModuleAttribute(
    Version('buildbot', 4, 3, 0),
    message='Use buildbot.schedulers.dependent.Dependent',
    moduleName='buildbot.schedulers.basic',
    name='Dependent',
)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

from twisted.internet import defer
from twisted.python import log

from buildbot.reporters.base import ReporterBase
from buildbot.reporters.generators.build import BuildStatusGenerator
from buildbot.reporters.message import MessageFormatterFunction
from buildbot.util import httpclientservice


class HttpStatusPush(ReporterBase):
    name: str | None = "HttpStatusPush"  # type: ignore[assignment]
    secrets = ["auth"]

    def checkConfig(
        self,
        serverUrl,
        auth=None,
        headers=None,
        debug=None,
        verify=None,
        cert=None,
        skip_encoding=False,
        generators=None,
        **kwargs,
    ):
        if generators is None:
            generators = self._create_default_generators()

        super().checkConfig(generators=generators, **kwargs)

    @defer.inlineCallbacks
    def reconfigService(
        self,
        serverUrl,
        auth=None,
        headers=None,
        debug=None,
        verify=None,
        cert=None,
        skip_encoding=False,
        generators=None,
        **kwargs,
    ):
        self.debug = debug
        self.verify = verify
        self.cert = cert

        if generators is None:
            generators = self._create_default_generators()

        yield super().reconfigService(generators=generators, **kwargs)

        self._http = yield httpclientservice.HTTPSession(
            self.master.httpservice,
            serverUrl,
            auth=auth,
            headers=headers,
            debug=self.debug,
            verify=self.verify,
            cert=self.cert,
            skip_encoding=skip_encoding,
        )

    def _create_default_generators(self):
        formatter = MessageFormatterFunction(lambda context: context['build'], 'json')
        return [BuildStatusGenerator(message_formatter=formatter, report_new=True)]

    def is_status_2xx(self, code):
        return code // 100 == 2

    @defer.inlineCallbacks
    def sendMessage(self, reports):
        response = yield self._http.post("", json=reports[0]['body'])
        if not self.is_status_2xx(response.code):
            log.msg(f"{response.code}: unable to upload status: {response.content}")



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import builtins
import errno
import os
import re
import shutil
import sys
from io import StringIO
from typing import TYPE_CHECKING
from unittest import mock

from twisted.python import log
from twisted.trial.unittest import TestCase

from buildbot_worker.scripts import base

if TYPE_CHECKING:
    from typing import Any
    from typing import Callable


def nl(s: str | Any) -> str | Any:
    """Convert the given string to the native newline format, assuming it is
    already in normal UNIX newline format (\n).  Use this to create the
    appropriate expectation in an assertEqual"""
    if not isinstance(s, str):
        return s
    return s.replace('\n', os.linesep)


class BasedirMixin:
    """Mix this in and call setUpBasedir and tearDownBasedir to set up
    a clean basedir with a name given in self.basedir."""

    def setUpBasedir(self) -> None:
        self.basedir = "test-basedir"
        if os.path.exists(self.basedir):
            shutil.rmtree(self.basedir)

    def tearDownBasedir(self) -> None:
        if os.path.exists(self.basedir):
            shutil.rmtree(self.basedir)


class IsWorkerDirMixin:
    """
    Mixin for setting up mocked base.isWorkerDir() function
    """

    def setupUpIsWorkerDir(self, return_value: bool) -> None:
        self.isWorkerDir = mock.Mock(return_value=return_value)
        assert isinstance(self, TestCase)
        self.patch(base, "isWorkerDir", self.isWorkerDir)


class PatcherMixin:
    """
    Mix this in to get a few special-cased patching methods
    """

    def patch_os_uname(self, replacement: Callable[[], os.uname_result]) -> None:
        # twisted's 'patch' doesn't handle the case where an attribute
        # doesn't exist..
        assert isinstance(self, TestCase)
        if hasattr(os, 'uname'):
            self.patch(os, 'uname', replacement)
        else:

            def cleanup() -> None:
                del os.uname

            self.addCleanup(cleanup)
            os.uname = replacement


class FileIOMixin:
    """
    Mixin for patching open(), read() and write() to simulate successful
    I/O operations and various I/O errors.
    """

    def setUpOpen(self, file_contents: str = "dummy-contents") -> None:
        """
        patch open() to return file object with provided contents.

        @param file_contents: contents that will be returned by file object's
                              read() method
        """
        assert isinstance(self, TestCase)
        # Use mock.mock_open() to create a substitute for
        # open().
        fakeOpen = mock.mock_open(read_data=file_contents)

        # When fakeOpen() is called, it returns a Mock
        # that has these methods: read(), write(), __enter__(), __exit__().
        # read() will always return the value of the 'file_contents variable.
        self.fileobj = fakeOpen()

        # patch open() to always return our Mock file object
        self.open = mock.Mock(return_value=self.fileobj)
        self.patch(builtins, "open", self.open)

    def setUpOpenError(
        self,
        errno: int = errno.ENOENT,
        strerror: str = "dummy-msg",
        filename: str = "dummy-file",
    ) -> None:
        """
        patch open() to raise IOError

        @param    errno: exception's errno value
        @param strerror: exception's strerror value
        @param filename: exception's filename value
        """
        assert isinstance(self, TestCase)
        # Use mock.mock_open() to create a substitute for
        # open().
        fakeOpen = mock.mock_open()

        # Add side_effect so that calling fakeOpen() will always
        # raise an IOError.
        fakeOpen.side_effect = OSError(errno, strerror, filename)
        self.open = fakeOpen
        self.patch(builtins, "open", self.open)

    def setUpReadError(
        self,
        errno: int = errno.EIO,
        strerror: str = "dummy-msg",
        filename: str = "dummy-file",
    ) -> None:
        """
        patch open() to return a file object that will raise IOError on read()

        @param    errno: exception's errno value
        @param strerror: exception's strerror value
        @param filename: exception's filename value

        """
        assert isinstance(self, TestCase)
        # Use mock.mock_open() to create a substitute for
        # open().
        fakeOpen = mock.mock_open()

        # When fakeOpen() is called, it returns a Mock
        # that has these methods: read(), write(), __enter__(), __exit__().
        self.fileobj = fakeOpen()

        # Add side_effect so that calling read() will always
        # raise an IOError.
        self.fileobj.read.side_effect = OSError(errno, strerror, filename)

        # patch open() to always return our Mock file object
        self.open = mock.Mock(return_value=self.fileobj)
        self.patch(builtins, "open", self.open)

    def setUpWriteError(
        self,
        errno: int = errno.ENOSPC,
        strerror: str = "dummy-msg",
        filename: str = "dummy-file",
    ) -> None:
        """
        patch open() to return a file object that will raise IOError on write()

        @param    errno: exception's errno value
        @param strerror: exception's strerror value
        @param filename: exception's filename value
        """
        assert isinstance(self, TestCase)
        # Use mock.mock_open() to create a substitute for
        # open().
        fakeOpen = mock.mock_open()

        # When fakeOpen() is called, it returns a Mock
        # that has these methods: read(), write(), __enter__(), __exit__().
        self.fileobj = fakeOpen()

        # Add side_effect so that calling write() will always
        # raise an IOError.
        self.fileobj.write.side_effect = OSError(errno, strerror, filename)

        # patch open() to always return our Mock file object
        self.open = mock.Mock(return_value=self.fileobj)
        self.patch(builtins, "open", self.open)


class LoggingMixin:
    def setUpLogging(self) -> None:
        assert isinstance(self, TestCase)
        self._logEvents: list[log.EventDict] = []
        log.addObserver(self._logEvents.append)
        self.addCleanup(log.removeObserver, self._logEvents.append)

    def assertLogged(self, *args: str) -> None:
        assert isinstance(self, TestCase)
        for regexp in args:
            r = re.compile(regexp)
            for event in self._logEvents:
                msg = log.textFromEventDict(event)
                if msg is not None and r.search(msg):
                    return
            self.fail(f"{regexp!r} not matched in log output.\n{self._logEvents} ")

    def assertWasQuiet(self) -> None:
        assert isinstance(self, TestCase)
        self.assertEqual(self._logEvents, [])


class StdoutAssertionsMixin:
    """
    Mix this in to be able to assert on stdout during the test
    """

    def setUpStdoutAssertions(self) -> None:
        assert isinstance(self, TestCase)
        self.stdout = StringIO()
        self.patch(sys, 'stdout', self.stdout)

    def assertWasQuiet(self) -> None:
        assert isinstance(self, TestCase)
        self.assertEqual(self.stdout.getvalue(), '')

    def assertInStdout(self, exp: str) -> None:
        assert isinstance(self, TestCase)
        self.assertIn(exp, self.stdout.getvalue())

    def assertStdoutEqual(self, exp: str, msg: str | None = None) -> None:
        assert isinstance(self, TestCase)
        self.assertEqual(exp, self.stdout.getvalue(), msg)

    def getStdout(self) -> str:
        return self.stdout.getvalue().strip()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from twisted.trial import unittest

from buildbot.process.results import SUCCESS
from buildbot.steps.package.rpm import rpmlint
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import TestBuildStepMixin


class TestRpmLint(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_success(self):
        self.setup_step(rpmlint.RpmLint())
        self.expect_commands(ExpectShell(workdir='wkdir', command=['rpmlint', '-i', '.']).exit(0))
        self.expect_outcome(result=SUCCESS, state_string='Finished checking RPM/SPEC issues')
        return self.run_step()

    def test_fileloc_success(self):
        self.setup_step(rpmlint.RpmLint(fileloc='RESULT'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['rpmlint', '-i', 'RESULT']).exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_config_success(self):
        self.setup_step(rpmlint.RpmLint(config='foo.cfg'))
        self.expect_commands(
            ExpectShell(workdir='wkdir', command=['rpmlint', '-i', '-f', 'foo.cfg', '.']).exit(0)
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import inspect
from functools import wraps
from typing import Any
from typing import Callable
from typing import Coroutine
from typing import Generator
from typing import TypeVar
from typing import Union

from twisted.internet import defer
from twisted.internet import reactor
from twisted.python import threadpool
from typing_extensions import ParamSpec

_T = TypeVar('_T')
_P = ParamSpec('_P')


InlineCallbacksType = Generator[Union[Any, defer.Deferred[Any]], Any, _T]


def async_to_deferred(
    fn: Callable[_P, Coroutine[Any, Any, _T]],
) -> Callable[_P, defer.Deferred[_T]]:
    @wraps(fn)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> defer.Deferred[_T]:
        try:
            return defer.ensureDeferred(fn(*args, **kwargs))
        except Exception as e:
            return defer.fail(e)

    return wrapper


async def any_to_async(value: Coroutine[Any, Any, _T] | defer.Deferred[_T] | _T) -> _T:
    if inspect.isawaitable(value):
        # defer.Deferred is awaitable too
        return await value
    return value


class ThreadPool(threadpool.ThreadPool):
    # This thread pool ensures that it stops on reactor shutdown

    _stop_event = None  # if not None, then pool is running

    def start(self):
        if self._stop_event:
            return

        super().start()
        self._stop_event = reactor.addSystemEventTrigger(
            'during', 'shutdown', self._stop_on_shutdown
        )

    def _stop_on_shutdown(self):
        self._stop_impl(remove_trigger=False)

    def stop(self):
        self._stop_impl(remove_trigger=True)

    def _stop_impl(self, remove_trigger):
        if not self._stop_event:
            return
        super().stop()
        if remove_trigger:
            reactor.removeSystemEventTrigger(self._stop_event)
        self._stop_event = None



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

import time

from twisted.trial import unittest

from buildbot import config
from buildbot.process.properties import Interpolate
from buildbot.process.results import FAILURE
from buildbot.process.results import SUCCESS
from buildbot.steps.package.deb import pbuilder
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.steps import ExpectShell
from buildbot.test.steps import ExpectStat
from buildbot.test.steps import TestBuildStepMixin


class TestDebPbuilder(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_new(self):
        self.setup_step(pbuilder.DebPbuilder())
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-local-buildbot.tgz').exit(1),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/pbuilder',
                    '--create',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                    '--distribution',
                    'stable',
                    '--mirror',
                    'http://cdn.debian.net/debian/',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/pbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS, state_string='built')
        return self.run_step()

    def test_update(self):
        self.setup_step(pbuilder.DebPbuilder())
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-local-buildbot.tgz').stat_file().exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/pbuilder',
                    '--update',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/pbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_buildonly_and_property(self):
        self.setup_step(pbuilder.DebPbuilder())
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-local-buildbot.tgz')
            .stat_file(mtime=int(time.time()))
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/pbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                ],
            )
            .stdout('blah\ndpkg-genchanges  >../somefilename.changes\foo\n')
            .exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        self.expect_property('deb-changes', 'somefilename.changes', 'DebPbuilder')
        return self.run_step()

    def test_architecture(self):
        self.setup_step(pbuilder.DebPbuilder(architecture='amd64'))
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-amd64-buildbot.tgz').exit(1),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/pbuilder',
                    '--create',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-amd64-buildbot.tgz',
                    '--distribution',
                    'stable',
                    '--mirror',
                    'http://cdn.debian.net/debian/',
                    '--architecture',
                    'amd64',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/pbuilder',
                    '--architecture',
                    'amd64',
                    '--',
                    '--buildresult',
                    '.',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-amd64-buildbot.tgz',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_architecture_renderable(self):
        self.setup_step(pbuilder.DebPbuilder(architecture=Interpolate('amd64')))
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-amd64-buildbot.tgz').exit(1),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/pbuilder',
                    '--create',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-amd64-buildbot.tgz',
                    '--distribution',
                    'stable',
                    '--mirror',
                    'http://cdn.debian.net/debian/',
                    '--architecture',
                    'amd64',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/pbuilder',
                    '--architecture',
                    'amd64',
                    '--',
                    '--buildresult',
                    '.',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-amd64-buildbot.tgz',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_distribution(self):
        self.setup_step(pbuilder.DebPbuilder(distribution='woody'))
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/woody-local-buildbot.tgz').exit(1),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/pbuilder',
                    '--create',
                    '--basetgz',
                    '/var/cache/pbuilder/woody-local-buildbot.tgz',
                    '--distribution',
                    'woody',
                    '--mirror',
                    'http://cdn.debian.net/debian/',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/pbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basetgz',
                    '/var/cache/pbuilder/woody-local-buildbot.tgz',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_basetgz(self):
        self.setup_step(pbuilder.DebPbuilder(basetgz='/buildbot/stable-local.tgz'))
        self.expect_commands(
            ExpectStat(file='/buildbot/stable-local.tgz').exit(1),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/pbuilder',
                    '--create',
                    '--basetgz',
                    '/buildbot/stable-local.tgz',
                    '--distribution',
                    'stable',
                    '--mirror',
                    'http://cdn.debian.net/debian/',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/pbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basetgz',
                    '/buildbot/stable-local.tgz',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_mirror(self):
        self.setup_step(pbuilder.DebPbuilder(mirror='http://apt:9999/debian'))
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-local-buildbot.tgz').exit(1),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/pbuilder',
                    '--create',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                    '--distribution',
                    'stable',
                    '--mirror',
                    'http://apt:9999/debian',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/pbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_extrapackages(self):
        self.setup_step(pbuilder.DebPbuilder(extrapackages=['buildbot']))
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-local-buildbot.tgz').exit(1),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/pbuilder',
                    '--create',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                    '--distribution',
                    'stable',
                    '--mirror',
                    'http://cdn.debian.net/debian/',
                    '--extrapackages',
                    'buildbot',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/pbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                    '--extrapackages',
                    'buildbot',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_keyring(self):
        self.setup_step(pbuilder.DebPbuilder(keyring='/builbot/buildbot.gpg'))
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-local-buildbot.tgz').exit(1),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/pbuilder',
                    '--create',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                    '--distribution',
                    'stable',
                    '--mirror',
                    'http://cdn.debian.net/debian/',
                    '--debootstrapopts',
                    '--keyring=/builbot/buildbot.gpg',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/pbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_components(self):
        self.setup_step(pbuilder.DebPbuilder(components='main universe'))
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-local-buildbot.tgz').exit(1),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/pbuilder',
                    '--create',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                    '--distribution',
                    'stable',
                    '--mirror',
                    'http://cdn.debian.net/debian/',
                    '--components',
                    'main universe',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/pbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_othermirror(self):
        self.setup_step(pbuilder.DebPbuilder(othermirror=['http://apt:9999/debian']))
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-local-buildbot.tgz').exit(1),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/pbuilder',
                    '--create',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                    '--distribution',
                    'stable',
                    '--mirror',
                    'http://cdn.debian.net/debian/',
                    '--othermirror',
                    'http://apt:9999/debian',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/pbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basetgz',
                    '/var/cache/pbuilder/stable-local-buildbot.tgz',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()


class TestDebCowbuilder(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_new(self):
        self.setup_step(pbuilder.DebCowbuilder())
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-local-buildbot.cow/').exit(1),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/cowbuilder',
                    '--create',
                    '--basepath',
                    '/var/cache/pbuilder/stable-local-buildbot.cow/',
                    '--distribution',
                    'stable',
                    '--mirror',
                    'http://cdn.debian.net/debian/',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/cowbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basepath',
                    '/var/cache/pbuilder/stable-local-buildbot.cow/',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_update(self):
        self.setup_step(pbuilder.DebCowbuilder())
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-local-buildbot.cow/').stat_dir().exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/cowbuilder',
                    '--update',
                    '--basepath',
                    '/var/cache/pbuilder/stable-local-buildbot.cow/',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/cowbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basepath',
                    '/var/cache/pbuilder/stable-local-buildbot.cow/',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_buildonly(self):
        self.setup_step(pbuilder.DebCowbuilder())
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-local-buildbot.cow/')
            .stat_dir(mtime=int(time.time()))
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/cowbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basepath',
                    '/var/cache/pbuilder/stable-local-buildbot.cow/',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS)
        return self.run_step()

    def test_update_reg(self):
        self.setup_step(
            pbuilder.DebCowbuilder(basetgz='/var/cache/pbuilder/stable-local-buildbot.cow')
        )
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-local-buildbot.cow').stat_file().exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/cowbuilder',
                    '--update',
                    '--basepath',
                    '/var/cache/pbuilder/stable-local-buildbot.cow',
                ],
            ).exit(1),
        )
        self.expect_outcome(result=FAILURE, state_string='built (failure)')
        return self.run_step()

    def test_buildonly_reg(self):
        self.setup_step(
            pbuilder.DebCowbuilder(basetgz='/var/cache/pbuilder/stable-local-buildbot.cow')
        )
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/stable-local-buildbot.cow')
            .stat_file(mtime=int(time.time()))
            .exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/cowbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basepath',
                    '/var/cache/pbuilder/stable-local-buildbot.cow',
                ],
            ).exit(1),
        )
        self.expect_outcome(result=FAILURE, state_string='built (failure)')
        return self.run_step()


class TestUbuPbuilder(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_no_distribution(self):
        with self.assertRaises(config.ConfigErrors):
            pbuilder.UbuPbuilder()

    def test_new(self):
        self.setup_step(pbuilder.UbuPbuilder(distribution='oneiric'))
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/oneiric-local-buildbot.tgz').exit(1),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/pbuilder',
                    '--create',
                    '--basetgz',
                    '/var/cache/pbuilder/oneiric-local-buildbot.tgz',
                    '--distribution',
                    'oneiric',
                    '--mirror',
                    'http://archive.ubuntu.com/ubuntu/',
                    '--components',
                    'main universe',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/pbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basetgz',
                    '/var/cache/pbuilder/oneiric-local-buildbot.tgz',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS, state_string='built')
        return self.run_step()


class TestUbuCowbuilder(TestBuildStepMixin, TestReactorMixin, unittest.TestCase):
    def setUp(self):
        self.setup_test_reactor()
        return self.setup_test_build_step()

    def test_no_distribution(self):
        with self.assertRaises(config.ConfigErrors):
            pbuilder.UbuCowbuilder()

    def test_new(self):
        self.setup_step(pbuilder.UbuCowbuilder(distribution='oneiric'))
        self.expect_commands(
            ExpectStat(file='/var/cache/pbuilder/oneiric-local-buildbot.cow/').exit(1),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'sudo',
                    '/usr/sbin/cowbuilder',
                    '--create',
                    '--basepath',
                    '/var/cache/pbuilder/oneiric-local-buildbot.cow/',
                    '--distribution',
                    'oneiric',
                    '--mirror',
                    'http://archive.ubuntu.com/ubuntu/',
                    '--components',
                    'main universe',
                ],
            ).exit(0),
            ExpectShell(
                workdir='wkdir',
                command=[
                    'pdebuild',
                    '--buildresult',
                    '.',
                    '--pbuilder',
                    '/usr/sbin/cowbuilder',
                    '--',
                    '--buildresult',
                    '.',
                    '--basepath',
                    '/var/cache/pbuilder/oneiric-local-buildbot.cow/',
                ],
            ).exit(0),
        )
        self.expect_outcome(result=SUCCESS, state_string='built')
        return self.run_step()



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from unittest.mock import Mock

from parameterized import parameterized
from twisted.internet import defer
from twisted.trial import unittest

from buildbot.process.results import SUCCESS
from buildbot.reporters import utils
from buildbot.reporters.generators.build import BuildStartEndStatusGenerator
from buildbot.reporters.generators.build import BuildStatusGenerator
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.util.config import ConfigErrorsMixin
from buildbot.test.util.reporter import ReporterTestMixin
from buildbot.test.util.warnings import assertProducesWarnings
from buildbot.warnings import DeprecatedApiWarning


class TestBuildGenerator(ConfigErrorsMixin, TestReactorMixin, unittest.TestCase, ReporterTestMixin):
    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.setup_reporter_test()
        self.master = yield fakemaster.make_master(self, wantData=True, wantDb=True, wantMq=True)

    @defer.inlineCallbacks
    def insert_build_finished_get_props(
        self, results, add_logs=None, want_logs_content=False, **kwargs
    ):
        build = yield self.insert_build_finished(results, **kwargs)
        yield utils.getDetailsForBuild(
            self.master,
            build,
            want_properties=True,
            add_logs=add_logs,
            want_logs_content=want_logs_content,
        )
        return build

    @defer.inlineCallbacks
    def setup_generator(
        self,
        results=SUCCESS,
        message=None,
        db_args=None,
        add_logs=None,
        want_logs_content=False,
        **kwargs,
    ):
        if message is None:
            message = {
                "body": "body",
                "type": "text",
                "subject": "subject",
                "extra_info": None,
            }
        if db_args is None:
            db_args = {}

        build = yield self.insert_build_finished_get_props(
            results, want_logs_content=want_logs_content, add_logs=add_logs, **db_args
        )
        buildset = yield self.get_inserted_buildset()

        g = BuildStatusGenerator(add_logs=add_logs, **kwargs)

        g.formatter = Mock(spec=g.formatter)
        g.formatter.want_logs_content = want_logs_content
        g.formatter.format_message_for_build.return_value = message

        return g, build, buildset

    @defer.inlineCallbacks
    def build_message(self, g, build, results=SUCCESS):
        reporter = Mock()
        reporter.getResponsibleUsersForBuild.return_value = []

        report = yield g.build_message(g.formatter, self.master, reporter, build)
        return report

    @defer.inlineCallbacks
    def generate(self, g, key, build):
        reporter = Mock()
        reporter.getResponsibleUsersForBuild.return_value = []

        report = yield g.generate(self.master, reporter, key, build)
        return report

    @defer.inlineCallbacks
    def test_build_message_nominal(self):
        g, build, buildset = yield self.setup_generator(mode=("change",))
        report = yield self.build_message(g, build)

        g.formatter.format_message_for_build.assert_called_with(
            self.master, build, is_buildset=False, mode=('change',), users=[]
        )

        self.assertEqual(
            report,
            {
                'body': 'body',
                'subject': 'subject',
                'type': 'text',
                "extra_info": None,
                'results': SUCCESS,
                'builds': [build],
                "buildset": buildset,
                'users': [],
                'patches': [],
                'logs': [],
            },
        )

    @defer.inlineCallbacks
    def test_build_message_no_result(self):
        g, build, buildset = yield self.setup_generator(results=None, mode=("change",))
        report = yield self.build_message(g, build, results=None)

        g.formatter.format_message_for_build.assert_called_with(
            self.master, build, is_buildset=False, mode=('change',), users=[]
        )

        self.assertEqual(
            report,
            {
                'body': 'body',
                'subject': 'subject',
                'type': 'text',
                "extra_info": None,
                'results': None,
                'builds': [build],
                "buildset": buildset,
                'users': [],
                'patches': [],
                'logs': [],
            },
        )

    @defer.inlineCallbacks
    def test_build_message_no_result_formatter_no_subject(self):
        message = {
            "body": "body",
            "type": "text",
            "subject": None,  # deprecated unspecified subject
            "extra_info": None,
        }

        g, build, buildset = yield self.setup_generator(
            results=None, message=message, mode=("change",)
        )
        report = yield self.build_message(g, build, results=None)

        g.formatter.format_message_for_build.assert_called_with(
            self.master, build, is_buildset=False, mode=('change',), users=[]
        )

        self.assertEqual(
            report,
            {
                'body': 'body',
                'subject': 'Buildbot not finished in Buildbot on Builder0',
                'type': 'text',
                "extra_info": None,
                'results': None,
                'builds': [build],
                "buildset": buildset,
                'users': [],
                'patches': [],
                'logs': [],
            },
        )

    @defer.inlineCallbacks
    def test_build_message_addLogs(self):
        with assertProducesWarnings(
            DeprecatedApiWarning,
            message_pattern=".*argument add_logs have been deprecated.*",
        ):
            g, build, _ = yield self.setup_generator(mode=("change",), add_logs=True)
        report = yield self.build_message(g, build)

        self.assertEqual(report['logs'][0]['logid'], 60)
        self.assertIn("log with", report['logs'][0]['content']['content'])

    @defer.inlineCallbacks
    def test_build_message_want_logs_content(self):
        g, build, _ = yield self.setup_generator(mode=("change",), want_logs_content=True)
        report = yield self.build_message(g, build)

        self.assertEqual(report['logs'][0]['logid'], 60)
        self.assertIn("log with", report['logs'][0]['content']['content'])

    @defer.inlineCallbacks
    def test_build_message_add_patch(self):
        g, build, _ = yield self.setup_generator(
            mode=("change",), add_patch=True, db_args={"insert_patch": True}
        )
        report = yield self.build_message(g, build)

        patch_dict = {
            'author': 'him@foo',
            'body': b'hello, world',
            'comment': 'foo',
            'level': 3,
            'patchid': 99,
            'subdir': '/foo',
        }
        self.assertEqual(report['patches'], [patch_dict])

    @defer.inlineCallbacks
    def test_build_message_add_patch_no_patch(self):
        g, build, _ = yield self.setup_generator(
            mode=("change",), add_patch=True, db_args={'insert_patch': False}
        )
        report = yield self.build_message(g, build)
        self.assertEqual(report['patches'], [])

    @defer.inlineCallbacks
    def test_generate_finished(self):
        g, build, buildset = yield self.setup_generator()
        report = yield self.generate(g, ('builds', 123, 'finished'), build)

        self.assertEqual(
            report,
            {
                'body': 'body',
                'subject': 'subject',
                'type': 'text',
                "extra_info": None,
                'results': SUCCESS,
                'builds': [build],
                "buildset": buildset,
                'users': [],
                'patches': [],
                'logs': [],
            },
        )

    @defer.inlineCallbacks
    def test_generate_finished_non_matching_builder(self):
        g, build, _ = yield self.setup_generator(builders=['non-matched'])
        report = yield self.generate(g, ('builds', 123, 'finished'), build)

        self.assertIsNone(report)

    @defer.inlineCallbacks
    def test_generate_finished_non_matching_result(self):
        g, build, _ = yield self.setup_generator(mode=('failing',))
        report = yield self.generate(g, ('builds', 123, 'finished'), build)

        self.assertIsNone(report)

    @defer.inlineCallbacks
    def test_generate_new(self):
        g, build, buildset = yield self.setup_generator(
            results=None, mode=("failing",), report_new=True
        )
        report = yield self.generate(g, ('builds', 123, 'new'), build)

        self.assertEqual(
            report,
            {
                'body': 'body',
                'subject': 'subject',
                'type': 'text',
                "extra_info": None,
                'results': None,
                'builds': [build],
                "buildset": buildset,
                'users': [],
                'patches': [],
                'logs': [],
            },
        )


class TestBuildStartEndGenerator(
    ConfigErrorsMixin, TestReactorMixin, unittest.TestCase, ReporterTestMixin
):
    all_messages = ('failing', 'passing', 'warnings', 'exception', 'cancelled')

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.setup_reporter_test()
        self.master = yield fakemaster.make_master(self, wantData=True, wantDb=True, wantMq=True)

    @defer.inlineCallbacks
    def insert_build_finished_get_props(
        self, results, add_logs=None, want_logs_content=False, **kwargs
    ):
        build = yield self.insert_build_finished(results, **kwargs)
        yield utils.getDetailsForBuild(
            self.master,
            build,
            want_properties=True,
            add_logs=add_logs,
            want_logs_content=want_logs_content,
        )
        return build

    @parameterized.expand([
        ('tags', 'tag'),
        ('tags', 1),
        ('builders', 'builder'),
        ('builders', 1),
        ('schedulers', 'scheduler'),
        ('schedulers', 1),
        ('branches', 'branch'),
        ('branches', 1),
    ])
    def test_list_params_check_raises(self, arg_name, arg_value):
        kwargs = {arg_name: arg_value}
        g = BuildStartEndStatusGenerator(**kwargs)
        with self.assertRaisesConfigError('must be a list or None'):
            g.check()

    def setup_generator(
        self,
        results=SUCCESS,
        start_message=None,
        end_message=None,
        want_logs_content=False,
        **kwargs,
    ):
        if start_message is None:
            start_message = {
                "body": "start body",
                "type": "plain",
                "subject": "start subject",
                "extra_info": None,
            }

        if end_message is None:
            end_message = {
                "body": "end body",
                "type": "plain",
                "subject": "end subject",
                "extra_info": None,
            }

        g = BuildStartEndStatusGenerator(**kwargs)

        g.start_formatter = Mock(spec=g.start_formatter)
        g.start_formatter.want_logs_content = want_logs_content
        g.start_formatter.format_message_for_build.return_value = start_message
        g.end_formatter = Mock(spec=g.end_formatter)
        g.end_formatter.want_logs_content = want_logs_content
        g.end_formatter.format_message_for_build.return_value = end_message

        return g

    @defer.inlineCallbacks
    def build_message(self, g, build, results=SUCCESS):
        reporter = Mock()
        reporter.getResponsibleUsersForBuild.return_value = []

        report = yield g.build_message(g.start_formatter, self.master, reporter, build)
        return report

    @defer.inlineCallbacks
    def generate(self, g, key, build):
        reporter = Mock()
        reporter.getResponsibleUsersForBuild.return_value = []

        report = yield g.generate(self.master, reporter, key, build)
        return report

    @defer.inlineCallbacks
    def test_build_message_start(self):
        g = yield self.setup_generator()
        build = yield self.insert_build_finished_get_props(SUCCESS)
        buildset = yield self.get_inserted_buildset()
        report = yield self.build_message(g, build)

        g.start_formatter.format_message_for_build.assert_called_with(
            self.master, build, is_buildset=False, mode=self.all_messages, users=[]
        )

        self.assertEqual(
            report,
            {
                'body': 'start body',
                'subject': 'start subject',
                'type': 'plain',
                "extra_info": None,
                'results': SUCCESS,
                'builds': [build],
                "buildset": buildset,
                'users': [],
                'patches': [],
                'logs': [],
            },
        )

    @defer.inlineCallbacks
    def test_build_message_start_no_result(self):
        g = yield self.setup_generator(results=None)
        build = yield self.insert_build_new()
        buildset = yield self.get_inserted_buildset()
        build["buildset"] = buildset
        report = yield self.build_message(g, build, results=None)

        g.start_formatter.format_message_for_build.assert_called_with(
            self.master, build, is_buildset=False, mode=self.all_messages, users=[]
        )

        self.assertEqual(
            report,
            {
                'body': 'start body',
                'subject': 'start subject',
                'type': 'plain',
                "extra_info": None,
                'results': None,
                'builds': [build],
                "buildset": buildset,
                'users': [],
                'patches': [],
                'logs': [],
            },
        )

    @defer.inlineCallbacks
    def test_is_message_needed_ignores_unspecified_tags(self):
        build = yield self.insert_build_finished_get_props(SUCCESS)

        # force tags
        build['builder']['tags'] = ['tag']
        g = BuildStartEndStatusGenerator(tags=['not_existing_tag'])
        self.assertFalse(g.is_message_needed_by_props(build))

    @defer.inlineCallbacks
    def test_is_message_needed_tags(self):
        build = yield self.insert_build_finished_get_props(SUCCESS)

        # force tags
        build['builder']['tags'] = ['tag']
        g = BuildStartEndStatusGenerator(tags=['tag'])
        self.assertTrue(g.is_message_needed_by_props(build))

    @defer.inlineCallbacks
    def test_build_message_add_logs(self):
        with assertProducesWarnings(
            DeprecatedApiWarning,
            message_pattern=".*argument add_logs have been deprecated.*",
        ):
            g = yield self.setup_generator(add_logs=True)
        build = yield self.insert_build_finished_get_props(SUCCESS, add_logs=True)
        report = yield self.build_message(g, build)

        self.assertEqual(report['logs'][0]['logid'], 60)
        self.assertIn("log with", report['logs'][0]['content']['content'])

    @defer.inlineCallbacks
    def test_build_message_want_logs_content(self):
        g = yield self.setup_generator(want_logs_content=True)
        build = yield self.insert_build_finished_get_props(SUCCESS, want_logs_content=True)
        report = yield self.build_message(g, build)

        self.assertEqual(report['logs'][0]['logid'], 60)
        self.assertIn("log with", report['logs'][0]['content']['content'])

    @defer.inlineCallbacks
    def test_build_message_add_patch(self):
        g = yield self.setup_generator(add_patch=True)
        build = yield self.insert_build_finished_get_props(SUCCESS, insert_patch=True)
        report = yield self.build_message(g, build)

        patch_dict = {
            'author': 'him@foo',
            'body': b'hello, world',
            'comment': 'foo',
            'level': 3,
            'patchid': 99,
            'subdir': '/foo',
        }
        self.assertEqual(report['patches'], [patch_dict])

    @defer.inlineCallbacks
    def test_build_message_add_patch_no_patch(self):
        g = yield self.setup_generator(add_patch=True)
        build = yield self.insert_build_finished_get_props(SUCCESS, insert_patch=False)
        report = yield self.build_message(g, build)
        self.assertEqual(report['patches'], [])

    @defer.inlineCallbacks
    def test_generate_new(self):
        g = yield self.setup_generator()
        build = yield self.insert_build_new()
        buildset = yield self.get_inserted_buildset()
        report = yield self.generate(g, ('builds', 123, 'new'), build)

        self.assertEqual(
            report,
            {
                'body': 'start body',
                'subject': 'start subject',
                'type': 'plain',
                "extra_info": None,
                'results': None,
                'builds': [build],
                "buildset": buildset,
                'users': [],
                'patches': [],
                'logs': [],
            },
        )

    @defer.inlineCallbacks
    def test_generate_finished(self):
        g = yield self.setup_generator()
        build = yield self.insert_build_finished_get_props(SUCCESS)
        buildset = yield self.get_inserted_buildset()
        report = yield self.generate(g, ('builds', 123, 'finished'), build)

        self.assertEqual(
            report,
            {
                'body': 'end body',
                'subject': 'end subject',
                'type': 'plain',
                "extra_info": None,
                'results': SUCCESS,
                'builds': [build],
                "buildset": buildset,
                'users': [],
                'patches': [],
                'logs': [],
            },
        )

    @defer.inlineCallbacks
    def test_generate_none(self):
        g = yield self.setup_generator(builders=['other builder'])
        build = yield self.insert_build_new()
        build["buildset"] = yield self.get_inserted_buildset()
        report = yield self.generate(g, ('builds', 123, 'new'), build)

        self.assertIsNone(report, None)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members
from __future__ import annotations

import os
import xml.dom.minidom

from twisted.internet import defer
from twisted.trial import unittest

from buildbot.changes import svnpoller
from buildbot.process.properties import Interpolate
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.runprocess import ExpectMasterShell
from buildbot.test.runprocess import MasterRunProcessMixin
from buildbot.test.util import changesource

# this is the output of "svn info --xml
# svn+ssh://svn.twistedmatrix.com/svn/Twisted/trunk"
prefix_output = b"""\
<?xml version="1.0"?>
<info>
<entry
   kind="dir"
   path="trunk"
   revision="18354">
<url>svn+ssh://svn.twistedmatrix.com/svn/Twisted/trunk</url>
<repository>
<root>svn+ssh://svn.twistedmatrix.com/svn/Twisted</root>
<uuid>bbbe8e31-12d6-0310-92fd-ac37d47ddeeb</uuid>
</repository>
<commit
   revision="18352">
<author>jml</author>
<date>2006-10-01T02:37:34.063255Z</date>
</commit>
</entry>
</info>
"""

# and this is "svn info --xml svn://svn.twistedmatrix.com/svn/Twisted". I
# think this is kind of a degenerate case.. it might even be a form of error.
prefix_output_2 = b"""\
<?xml version="1.0"?>
<info>
</info>
"""

# this is the svn info output for a local repository, svn info --xml
# file:///home/warner/stuff/Projects/Buildbot/trees/svnpoller/_trial_temp/test_vc/repositories/SVN-Repository
prefix_output_3 = b"""\
<?xml version="1.0"?>
<info>
<entry
   kind="dir"
   path="SVN-Repository"
   revision="3">
<url>file:///home/warner/stuff/Projects/Buildbot/trees/svnpoller/_trial_temp/test_vc/repositories/SVN-Repository</url>
<repository>
<root>file:///home/warner/stuff/Projects/Buildbot/trees/svnpoller/_trial_temp/test_vc/repositories/SVN-Repository</root>
<uuid>c0f47ff4-ba1e-0410-96b5-d44cc5c79e7f</uuid>
</repository>
<commit
   revision="3">
<author>warner</author>
<date>2006-10-01T07:37:04.182499Z</date>
</commit>
</entry>
</info>
"""

# % svn info --xml file:///home/warner/stuff/Projects/Buildbot/trees/svnpoller/_trial_temp/test_vc/repositories/SVN-Repository/sample/trunk

prefix_output_4 = b"""\
<?xml version="1.0"?>
<info>
<entry
   kind="dir"
   path="trunk"
   revision="3">
<url>file:///home/warner/stuff/Projects/Buildbot/trees/svnpoller/_trial_temp/test_vc/repositories/SVN-Repository/sample/trunk</url>
<repository>
<root>file:///home/warner/stuff/Projects/Buildbot/trees/svnpoller/_trial_temp/test_vc/repositories/SVN-Repository</root>
<uuid>c0f47ff4-ba1e-0410-96b5-d44cc5c79e7f</uuid>
</repository>
<commit
   revision="1">
<author>warner</author>
<date>2006-10-01T07:37:02.286440Z</date>
</commit>
</entry>
</info>
"""


# output from svn log on .../SVN-Repository/sample
# (so it includes trunk and branches)
sample_base = (
    "file:///usr/home/warner/stuff/Projects/Buildbot/trees/misc/"
    + "_trial_temp/test_vc/repositories/SVN-Repository/sample"
)
sample_logentries: list[bytes | None] = [None] * 6

sample_logentries[5] = b"""\
<logentry
   revision="6">
<author>warner</author>
<date>2006-10-01T19:35:16.165664Z</date>
<paths>
<path
   action="D">/sample/branch/version.c</path>
</paths>
<msg>revised_to_2</msg>
</logentry>
"""

sample_logentries[4] = b"""\
<logentry
   revision="5">
<author>warner</author>
<date>2006-10-01T19:35:16.165664Z</date>
<paths>
<path
   action="D">/sample/branch</path>
</paths>
<msg>revised_to_2</msg>
</logentry>
"""

sample_logentries[3] = b"""\
<logentry
   revision="4">
<author>warner</author>
<date>2006-10-01T19:35:16.165664Z</date>
<paths>
<path
   action="M">/sample/trunk/version.c</path>
</paths>
<msg>revised_to_2</msg>
</logentry>
"""

sample_logentries[2] = b"""\
<logentry
   revision="3">
<author>warner</author>
<date>2006-10-01T19:35:10.215692Z</date>
<paths>
<path
   action="M">/sample/branch/c\xcc\xa7main.c</path>
</paths>
<msg>commit_on_branch</msg>
</logentry>
"""

sample_logentries[1] = b"""\
<logentry
   revision="2">
<author>warner</author>
<date>2006-10-01T19:35:09.154973Z</date>
<paths>
<path
   copyfrom-path="/sample/trunk"
   copyfrom-rev="1"
   action="A">/sample/branch</path>
</paths>
<msg>make_branch</msg>
</logentry>
"""

sample_logentries[0] = b"""\
<logentry
   revision="1">
<author>warner</author>
<date>2006-10-01T19:35:08.642045Z</date>
<paths>
<path
   action="A">/sample</path>
<path
   action="A">/sample/trunk</path>
<path
   action="A">/sample/trunk/subdir/subdir.c</path>
<path
   action="A">/sample/trunk/main.c</path>
<path
   action="A">/sample/trunk/version.c</path>
<path
   action="A">/sample/trunk/subdir</path>
</paths>
<msg>sample_project_files</msg>
</logentry>
"""

sample_info_output = b"""\
<?xml version="1.0"?>
<info>
<entry
   kind="dir"
   path="sample"
   revision="4">
<url>file:///usr/home/warner/stuff/Projects/Buildbot/trees/misc/_trial_temp/test_vc/repositories/SVN-Repository/sample</url>
<repository>
<root>file:///usr/home/warner/stuff/Projects/Buildbot/trees/misc/_trial_temp/test_vc/repositories/SVN-Repository</root>
<uuid>4f94adfc-c41e-0410-92d5-fbf86b7c7689</uuid>
</repository>
<commit
   revision="4">
<author>warner</author>
<date>2006-10-01T19:35:16.165664Z</date>
</commit>
</entry>
</info>
"""


def make_changes_output(maxrevision):
    # return what 'svn log' would have just after the given revision was
    # committed
    logs = sample_logentries[0:maxrevision]
    assert len(logs) == maxrevision
    logs.reverse()
    output = (
        b"""<?xml version="1.0"?>
<log>"""
        + b"".join(logs)
        + b"</log>"
    )
    return output


def make_logentry_elements(maxrevision):
    "return the corresponding logentry elements for the given revisions"
    doc = xml.dom.minidom.parseString(make_changes_output(maxrevision))
    return doc.getElementsByTagName("logentry")


def split_file(path):
    pieces = path.split("/")
    if pieces[0] == "branch":
        return {"branch": 'branch', "path": '/'.join(pieces[1:])}
    if pieces[0] == "trunk":
        return {"path": '/'.join(pieces[1:])}
    raise RuntimeError(f"there shouldn't be any files like {path!r}")


class TestSVNPoller(
    MasterRunProcessMixin, changesource.ChangeSourceMixin, TestReactorMixin, unittest.TestCase
):
    def setUp(self):
        self.setup_test_reactor()
        self.setup_master_run_process()
        return self.setUpChangeSource()

    @defer.inlineCallbacks
    def attachSVNPoller(self, *args, **kwargs):
        s = svnpoller.SVNPoller(*args, **kwargs)
        yield self.attachChangeSource(s)
        return s

    @defer.inlineCallbacks
    def test_describe(self):
        s = yield self.attachSVNPoller('file://')
        self.assertSubstring("SVNPoller", s.describe())

    @defer.inlineCallbacks
    def test_name(self):
        s = yield self.attachSVNPoller('file://')
        self.assertEqual("file://", s.name)

        s = yield self.attachSVNPoller('file://', name='MyName')
        self.assertEqual("MyName", s.name)

    @defer.inlineCallbacks
    def test_strip_repourl(self):
        base = "svn+ssh://svn.twistedmatrix.com/svn/Twisted/trunk"
        s = yield self.attachSVNPoller(base + "/")
        self.assertEqual(s.repourl, base)

    @defer.inlineCallbacks
    def do_test_get_prefix(self, base, output, expected):
        s = yield self.attachSVNPoller(base)
        self.expect_commands(
            ExpectMasterShell(['svn', 'info', '--xml', '--non-interactive', base]).stdout(output)
        )
        prefix = yield s.get_prefix()

        self.assertEqual(prefix, expected)
        self.assert_all_commands_ran()

    def test_get_prefix_1(self):
        base = "svn+ssh://svn.twistedmatrix.com/svn/Twisted/trunk"
        return self.do_test_get_prefix(base, prefix_output, 'trunk')

    def test_get_prefix_2(self):
        base = "svn+ssh://svn.twistedmatrix.com/svn/Twisted"
        return self.do_test_get_prefix(base, prefix_output_2, '')

    def test_get_prefix_3(self):
        base = (
            "file:///home/warner/stuff/Projects/Buildbot/trees/"
            + "svnpoller/_trial_temp/test_vc/repositories/SVN-Repository"
        )
        return self.do_test_get_prefix(base, prefix_output_3, '')

    def test_get_prefix_4(self):
        base = (
            "file:///home/warner/stuff/Projects/Buildbot/trees/"
            + "svnpoller/_trial_temp/test_vc/repositories/SVN-Repository/sample/trunk"
        )
        return self.do_test_get_prefix(base, prefix_output_3, 'sample/trunk')

    @defer.inlineCallbacks
    def test_log_parsing(self):
        s = yield self.attachSVNPoller('file:///foo')
        output = make_changes_output(4)
        entries = s.parse_logs(output)
        # no need for elaborate assertions here; this is minidom's logic
        self.assertEqual(len(entries), 4)

    @defer.inlineCallbacks
    def test_get_new_logentries(self):
        s = yield self.attachSVNPoller('file:///foo')
        entries = make_logentry_elements(4)

        s.last_change = 4
        new = s.get_new_logentries(entries)
        self.assertEqual(s.last_change, 4)
        self.assertEqual(len(new), 0)

        s.last_change = 3
        new = s.get_new_logentries(entries)
        self.assertEqual(s.last_change, 4)
        self.assertEqual(len(new), 1)

        s.last_change = 1
        new = s.get_new_logentries(entries)
        self.assertEqual(s.last_change, 4)
        self.assertEqual(len(new), 3)

        # special case: if last_change is None, then no new changes are queued
        s.last_change = None
        new = s.get_new_logentries(entries)
        self.assertEqual(s.last_change, 4)
        self.assertEqual(len(new), 0)

    @defer.inlineCallbacks
    def test_get_text(self):
        doc = xml.dom.minidom.parseString(
            """
            <parent>
                <child>
                    hi
                    <grandchild>1</grandchild>
                    <grandchild>2</grandchild>
                </child>
            </parent>""".strip()
        )
        s = yield self.attachSVNPoller('http://', split_file=split_file)

        self.assertEqual(s._get_text(doc, 'grandchild'), '1')
        self.assertEqual(s._get_text(doc, 'nonexistent'), 'unknown')

    @defer.inlineCallbacks
    def test_create_changes(self):
        base = (
            "file:///home/warner/stuff/Projects/Buildbot/trees/"
            + "svnpoller/_trial_temp/test_vc/repositories/SVN-Repository/sample"
        )
        s = yield self.attachSVNPoller(base, split_file=split_file)
        s._prefix = "sample"

        logentries = dict(zip(range(1, 7), reversed(make_logentry_elements(6))))
        changes = s.create_changes(reversed([logentries[3], logentries[2]]))
        self.assertEqual(len(changes), 2)
        # note that parsing occurs in reverse
        self.assertEqual(changes[0]['branch'], "branch")
        self.assertEqual(changes[0]['revision'], '2')
        self.assertEqual(changes[0]['project'], '')
        self.assertEqual(changes[0]['repository'], base)
        self.assertEqual(changes[1]['branch'], "branch")
        self.assertEqual(changes[1]['files'], ["çmain.c"])
        self.assertEqual(changes[1]['revision'], '3')
        self.assertEqual(changes[1]['project'], '')
        self.assertEqual(changes[1]['repository'], base)

        changes = s.create_changes([logentries[4]])
        self.assertEqual(len(changes), 1)
        self.assertEqual(changes[0]['branch'], None)
        self.assertEqual(changes[0]['revision'], '4')
        self.assertEqual(changes[0]['files'], ["version.c"])

        # r5 should *not* create a change as it's a branch deletion
        changes = s.create_changes([logentries[5]])
        self.assertEqual(len(changes), 0)

        # r6 should create a change as it's not deleting an entire branch
        changes = s.create_changes([logentries[6]])
        self.assertEqual(len(changes), 1)
        self.assertEqual(changes[0]['branch'], 'branch')
        self.assertEqual(changes[0]['revision'], '6')
        self.assertEqual(changes[0]['files'], ["version.c"])

    def makeInfoExpect(self, password='bbrocks'):
        args = ['svn', 'info', '--xml', '--non-interactive', sample_base, '--username=dustin']
        if password is not None:
            args.append('--password=' + password)
        return ExpectMasterShell(args)

    def makeLogExpect(self, password='bbrocks'):
        args = ['svn', 'log', '--xml', '--verbose', '--non-interactive', '--username=dustin']
        if password is not None:
            args.append('--password=' + password)
        args.extend(['--limit=100', sample_base])
        return ExpectMasterShell(args)

    @defer.inlineCallbacks
    def test_create_changes_overridden_project(self):
        def custom_split_file(path):
            f = split_file(path)
            if f:
                f["project"] = "overridden-project"
                f["repository"] = "overridden-repository"
                f["codebase"] = "overridden-codebase"
            return f

        base = (
            "file:///home/warner/stuff/Projects/Buildbot/trees/"
            + "svnpoller/_trial_temp/test_vc/repositories/SVN-Repository/sample"
        )
        s = yield self.attachSVNPoller(base, split_file=custom_split_file)
        s._prefix = "sample"

        logentries = dict(zip(range(1, 7), reversed(make_logentry_elements(6))))
        changes = s.create_changes(reversed([logentries[3], logentries[2]]))
        self.assertEqual(len(changes), 2)

        # note that parsing occurs in reverse
        self.assertEqual(changes[0]['branch'], "branch")
        self.assertEqual(changes[0]['revision'], '2')
        self.assertEqual(changes[0]['project'], "overridden-project")
        self.assertEqual(changes[0]['repository'], "overridden-repository")
        self.assertEqual(changes[0]['codebase'], "overridden-codebase")

        self.assertEqual(changes[1]['branch'], "branch")
        self.assertEqual(changes[1]['files'], ['çmain.c'])
        self.assertEqual(changes[1]['revision'], '3')
        self.assertEqual(changes[1]['project'], "overridden-project")
        self.assertEqual(changes[1]['repository'], "overridden-repository")
        self.assertEqual(changes[1]['codebase'], "overridden-codebase")

    @defer.inlineCallbacks
    def test_poll(self):
        s = yield self.attachSVNPoller(
            sample_base, split_file=split_file, svnuser='dustin', svnpasswd='bbrocks'
        )

        self.expect_commands(
            self.makeInfoExpect().stdout(sample_info_output),
            self.makeLogExpect().stdout(make_changes_output(1)),
            self.makeLogExpect().stdout(make_changes_output(1)),
            self.makeLogExpect().stdout(make_changes_output(2)),
            self.makeLogExpect().stdout(make_changes_output(4)),
        )
        # fire it the first time; it should do nothing
        yield s.poll()

        # no changes generated on the first iteration
        self.assertEqual(self.master.data.updates.changesAdded, [])
        self.assertEqual(s.last_change, 1)

        # now fire it again, nothing changing
        yield s.poll()

        self.assertEqual(self.master.data.updates.changesAdded, [])
        self.assertEqual(s.last_change, 1)

        # and again, with r2 this time
        yield s.poll()

        self.assertEqual(
            self.master.data.updates.changesAdded,
            [
                {
                    'author': 'warner',
                    'committer': None,
                    'branch': 'branch',
                    'category': None,
                    'codebase': None,
                    'comments': 'make_branch',
                    'files': [''],
                    'project': '',
                    'properties': None,
                    'repository': 'file:///usr/home/warner/stuff/Projects/Buildbot/trees/misc/_trial_temp/test_vc/repositories/SVN-Repository/sample',
                    'revision': '2',
                    'revlink': '',
                    'src': 'svn',
                    'when_timestamp': None,
                }
            ],
        )
        self.assertEqual(s.last_change, 2)

        # and again with both r3 and r4 appearing together
        self.master.data.updates.changesAdded = []
        yield s.poll()

        self.assertEqual(
            self.master.data.updates.changesAdded,
            [
                {
                    'author': 'warner',
                    'committer': None,
                    'branch': 'branch',
                    'category': None,
                    'codebase': None,
                    'comments': 'commit_on_branch',
                    'files': ['çmain.c'],
                    'project': '',
                    'properties': None,
                    'repository': 'file:///usr/home/warner/stuff/Projects/Buildbot/trees/misc/_trial_temp/test_vc/repositories/SVN-Repository/sample',
                    'revision': '3',
                    'revlink': '',
                    'src': 'svn',
                    'when_timestamp': None,
                },
                {
                    'author': 'warner',
                    'committer': None,
                    'branch': None,
                    'category': None,
                    'codebase': None,
                    'comments': 'revised_to_2',
                    'files': ['version.c'],
                    'project': '',
                    'properties': None,
                    'repository': 'file:///usr/home/warner/stuff/Projects/Buildbot/trees/misc/_trial_temp/test_vc/repositories/SVN-Repository/sample',
                    'revision': '4',
                    'revlink': '',
                    'src': 'svn',
                    'when_timestamp': None,
                },
            ],
        )
        self.assertEqual(s.last_change, 4)
        self.assert_all_commands_ran()

    @defer.inlineCallbacks
    def test_poll_empty_password(self):
        s = yield self.attachSVNPoller(
            sample_base, split_file=split_file, svnuser='dustin', svnpasswd=''
        )

        self.expect_commands(
            self.makeInfoExpect(password="").stdout(sample_info_output),
            self.makeLogExpect(password="").stdout(make_changes_output(1)),
            self.makeLogExpect(password="").stdout(make_changes_output(1)),
            self.makeLogExpect(password="").stdout(make_changes_output(2)),
            self.makeLogExpect(password="").stdout(make_changes_output(4)),
        )
        yield s.poll()

    @defer.inlineCallbacks
    def test_poll_no_password(self):
        s = yield self.attachSVNPoller(sample_base, split_file=split_file, svnuser='dustin')

        self.expect_commands(
            self.makeInfoExpect(password=None).stdout(sample_info_output),
            self.makeLogExpect(password=None).stdout(make_changes_output(1)),
            self.makeLogExpect(password=None).stdout(make_changes_output(1)),
            self.makeLogExpect(password=None).stdout(make_changes_output(2)),
            self.makeLogExpect(password=None).stdout(make_changes_output(4)),
        )
        yield s.poll()

    @defer.inlineCallbacks
    def test_poll_interpolated_password(self):
        s = yield self.attachSVNPoller(
            sample_base, split_file=split_file, svnuser='dustin', svnpasswd=Interpolate('pa$$')
        )

        self.expect_commands(
            self.makeInfoExpect(password='pa$$').stdout(sample_info_output),
            self.makeLogExpect(password='pa$$').stdout(make_changes_output(1)),
            self.makeLogExpect(password='pa$$').stdout(make_changes_output(1)),
            self.makeLogExpect(password='pa$$').stdout(make_changes_output(2)),
            self.makeLogExpect(password='pa$$').stdout(make_changes_output(4)),
        )
        yield s.poll()

    @defer.inlineCallbacks
    def test_poll_get_prefix_exception(self):
        s = yield self.attachSVNPoller(
            sample_base, split_file=split_file, svnuser='dustin', svnpasswd='bbrocks'
        )

        self.expect_commands(self.makeInfoExpect().stderr(b"error"))
        yield s.poll()

        # should have logged the RuntimeError, but not errback'd from poll
        self.assertEqual(len(self.flushLoggedErrors(EnvironmentError)), 1)
        self.assert_all_commands_ran()

    @defer.inlineCallbacks
    def test_poll_get_logs_exception(self):
        s = yield self.attachSVNPoller(
            sample_base, split_file=split_file, svnuser='dustin', svnpasswd='bbrocks'
        )
        s._prefix = "abc"  # skip the get_prefix stuff

        self.expect_commands(self.makeLogExpect().stderr(b"some error"))
        yield s.poll()

        # should have logged the RuntimeError, but not errback'd from poll
        self.assertEqual(len(self.flushLoggedErrors(EnvironmentError)), 1)
        self.assert_all_commands_ran()

    @defer.inlineCallbacks
    def test_cachepath_empty(self):
        cachepath = os.path.abspath('revcache')
        if os.path.exists(cachepath):
            os.unlink(cachepath)
        s = yield self.attachSVNPoller(sample_base, cachepath=cachepath)
        self.assertEqual(s.last_change, None)

    @defer.inlineCallbacks
    def test_cachepath_full(self):
        cachepath = os.path.abspath('revcache')
        with open(cachepath, "w", encoding='utf-8') as f:
            f.write('33')
        s = yield self.attachSVNPoller(sample_base, cachepath=cachepath)
        self.assertEqual(s.last_change, 33)

        s.last_change = 44
        s.finished_ok()
        with open(cachepath, encoding='utf-8') as f:
            self.assertEqual(f.read().strip(), '44')

    @defer.inlineCallbacks
    def test_cachepath_bogus(self):
        cachepath = os.path.abspath('revcache')
        with open(cachepath, "w", encoding='utf-8') as f:
            f.write('nine')
        s = yield self.attachSVNPoller(sample_base, cachepath=cachepath)
        self.assertEqual(s.last_change, None)
        self.assertEqual(s.cachepath, None)
        # it should have called log.err once with a ValueError
        self.assertEqual(len(self.flushLoggedErrors(ValueError)), 1)

    def test_constructor_pollInterval(self):
        return self.attachSVNPoller(sample_base, pollInterval=100)  # just don't fail!

    @defer.inlineCallbacks
    def test_extra_args(self):
        extra_args = [
            '--no-auth-cache',
        ]
        base = "svn+ssh://svn.twistedmatrix.com/svn/Twisted/trunk"

        s = yield self.attachSVNPoller(repourl=base, extra_args=extra_args)
        self.assertEqual(s.extra_args, extra_args)

    @defer.inlineCallbacks
    def test_use_svnurl(self):
        base = "svn+ssh://svn.twistedmatrix.com/svn/Twisted/trunk"
        with self.assertRaises(TypeError):
            yield self.attachSVNPoller(svnurl=base)


class TestSplitFile(unittest.TestCase):
    def test_split_file_alwaystrunk(self):
        self.assertEqual(svnpoller.split_file_alwaystrunk('foo'), {"path": 'foo'})

    def test_split_file_branches_trunk(self):
        self.assertEqual(svnpoller.split_file_branches('trunk/'), (None, ''))

    def test_split_file_branches_trunk_subdir(self):
        self.assertEqual(svnpoller.split_file_branches('trunk/subdir/'), (None, 'subdir/'))

    def test_split_file_branches_trunk_subfile(self):
        self.assertEqual(
            svnpoller.split_file_branches('trunk/subdir/file.c'), (None, 'subdir/file.c')
        )

    def test_split_file_branches_trunk_invalid(self):
        # file named trunk (not a directory):
        self.assertEqual(svnpoller.split_file_branches('trunk'), None)

    def test_split_file_branches_branch(self):
        self.assertEqual(svnpoller.split_file_branches('branches/1.5.x/'), ('branches/1.5.x', ''))

    def test_split_file_branches_branch_subdir(self):
        self.assertEqual(
            svnpoller.split_file_branches('branches/1.5.x/subdir/'), ('branches/1.5.x', 'subdir/')
        )

    def test_split_file_branches_branch_subfile(self):
        self.assertEqual(
            svnpoller.split_file_branches('branches/1.5.x/subdir/file.c'),
            ('branches/1.5.x', 'subdir/file.c'),
        )

    def test_split_file_branches_branch_invalid(self):
        # file named branches/1.5.x (not a directory):
        self.assertEqual(svnpoller.split_file_branches('branches/1.5.x'), None)

    def test_split_file_branches_otherdir(self):
        # other dirs are ignored:
        self.assertEqual(svnpoller.split_file_branches('tags/testthis/subdir/'), None)

    def test_split_file_branches_otherfile(self):
        # other files are ignored:
        self.assertEqual(svnpoller.split_file_branches('tags/testthis/subdir/file.c'), None)

    def test_split_file_projects_branches(self):
        self.assertEqual(
            svnpoller.split_file_projects_branches('buildbot/trunk/subdir/file.c'),
            {"project": 'buildbot', "path": 'subdir/file.c'},
        )
        self.assertEqual(
            svnpoller.split_file_projects_branches('buildbot/branches/1.5.x/subdir/file.c'),
            {"project": 'buildbot', "branch": 'branches/1.5.x', "path": 'subdir/file.c'},
        )
        # tags are ignored:
        self.assertEqual(
            svnpoller.split_file_projects_branches('buildbot/tags/testthis/subdir/file.c'), None
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members

from __future__ import annotations

import json
import platform
import sys
from unittest.case import SkipTest

from twisted.internet import defer
from twisted.internet import reactor
from twisted.trial import unittest

from buildbot.plugins.db import get_plugins
from buildbot.process.results import SUCCESS
from buildbot.process.results import WARNINGS
from buildbot.reporters import telegram
from buildbot.reporters import words
from buildbot.schedulers import forcesched
from buildbot.test import fakedb
from buildbot.test.fake import fakemaster
from buildbot.test.fake import httpclientservice as fakehttpclientservice
from buildbot.test.fake.web import FakeRequest
from buildbot.test.reactor import TestReactorMixin
from buildbot.test.unit.reporters.test_words import ContactMixin
from buildbot.util import httpclientservice
from buildbot.util import service
from buildbot.util import unicode2bytes


class FakeChannel(service.AsyncService):
    pass


class FakeContact:
    def __init__(self, user=None, channel=None):
        super().__init__()
        self.user_id = user['id']
        self.user_info = user
        self.channel = FakeChannel
        self.channel.chat_info = channel.chat_info
        self.template = None
        self.messages = []

    def handleMessage(self, message, **kwargs):
        self.messages.append(message)
        return defer.succeed(message)


class TestTelegramContact(ContactMixin, unittest.TestCase):
    channelClass = telegram.TelegramChannel
    contactClass = telegram.TelegramContact

    class botClass(words.StatusBot):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            self.query_cache = {}

        def send_message(self, chat, message, **kwargs):
            return {'message_id': 123}

        def edit_message(self, chat, msgid, message, **kwargs):
            return {'message_id': 123}

        def delete_message(self, chat, msgid):
            pass

        def send_sticker(self, chat, sticker, **kwargs):
            pass

        def edit_keyboard(self, chat, msg, keyboard=None):
            pass

        def getChannel(self, channel):
            return self.channelClass(self, channel)

        def post(self, path, **kwargs):
            return True

    USER: dict[str, str | int] = {
        "id": 123456789,
        "first_name": "Harry",
        "last_name": "Potter",
        "username": "harrypotter",
    }

    CHANNEL = {"id": -12345678, "title": "Hogwards", "type": "group"}

    PRIVATE = {"id": 123456789, "type": "private"}

    def patch_send(self):
        self.sent = []
        self.stickers = 0

        def send_message(chat, message, **kwargs):
            self.sent.append((chat, message, kwargs))
            return {'message_id': 123}

        self.bot.send_message = send_message

        def send_sticker(chat, sticker, **kwargs):
            self.stickers += 1

        self.bot.send_sticker = send_sticker

    @defer.inlineCallbacks
    def setUp(self):
        ContactMixin.setUp(self)
        self.contact1 = self.contactClass(
            user=self.USER, channel=self.channelClass(self.bot, self.PRIVATE)
        )
        yield self.contact1.channel.setServiceParent(self.master)

    @defer.inlineCallbacks
    def test_list_notified_events(self):
        self.patch_send()
        channel = telegram.TelegramChannel(self.bot, self.CHANNEL)
        channel.notify_events = {'success'}
        yield channel.list_notified_events()
        self.assertEqual(
            self.sent[0][1], "The following events are being notified:\n🔔 **success**"
        )

    @defer.inlineCallbacks
    def test_list_notified_events_empty(self):
        self.patch_send()
        channel = telegram.TelegramChannel(self.bot, self.CHANNEL)
        channel.notify_events = set()
        yield channel.list_notified_events()
        self.assertEqual(self.sent[0][1], "🔕 No events are being notified.")

    def testDescribeUser(self):
        self.assertEqual(self.contact1.describeUser(), "Harry Potter (@harrypotter)")

    def testDescribeUserInGroup(self):
        self.assertEqual(self.contact.describeUser(), "Harry Potter (@harrypotter) on 'Hogwards'")

    @defer.inlineCallbacks
    def test_access_denied(self):
        self.patch_send()
        self.contact1.ACCESS_DENIED_MESSAGES = ["ACCESS DENIED"]
        yield self.contact1.access_denied(tmessage={'message_id': 123})
        self.assertEqual("ACCESS DENIED", self.sent[0][1])

    @defer.inlineCallbacks
    def test_access_denied_group(self):
        self.patch_send()
        self.contact.ACCESS_DENIED_MESSAGES = ["ACCESS DENIED"]
        yield self.contact.access_denied(tmessage={'message_id': 123})
        self.assertEqual("ACCESS DENIED", self.sent[0][1])

    def test_query_button_short(self):
        result = self.contact.query_button("Hello", "hello")
        self.assertEqual(result, {'text': "Hello", 'callback_data': "hello"})

    def test_query_button_long(self):
        payload = 16 * "1234567890"
        key = hash(repr(payload))
        result = self.contact.query_button("Hello", payload)
        self.assertEqual(result, {'text': "Hello", 'callback_data': key})
        self.assertEqual(self.bot.query_cache[key], payload)

    def test_query_button_non_str(self):
        payload = {'data': "good"}
        key = hash(repr(payload))
        result = self.contact.query_button("Hello", payload)
        self.assertEqual(result, {'text': "Hello", 'callback_data': key})
        self.assertEqual(self.bot.query_cache[key], payload)

    def test_query_button_cache(self):
        payload = 16 * "1234567890"
        key = hash(repr(payload))
        self.bot.query_cache[key] = payload
        result = self.contact.query_button("Hello", payload)
        self.assertEqual(result, {'text': "Hello", 'callback_data': key})
        self.assertEqual(len(self.bot.query_cache), 1)

    def test_query_button_cache_conflict(self):
        payload = 16 * "1234567890"
        key = hash(repr(payload))
        self.bot.query_cache[key] = "something other"
        result = self.contact.query_button("Hello", payload)
        self.assertEqual(result, {'text': "Hello", 'callback_data': key + 1})
        self.assertEqual(self.bot.query_cache[key + 1], payload)

    @defer.inlineCallbacks
    def test_command_start(self):
        yield self.do_test_command('start', exp_usage=False)
        self.assertEqual(self.sent[0][0], self.CHANNEL['id'])

    @defer.inlineCallbacks
    def test_command_nay(self):
        yield self.do_test_command('nay', contact=self.contact1, tmessage={})

    @defer.inlineCallbacks
    def test_command_nay_reply_markup(self):
        yield self.do_test_command(
            'nay',
            tmessage={
                'reply_to_message': {
                    'message_id': 1234,
                    'reply_markup': {},
                }
            },
        )

    @defer.inlineCallbacks
    def test_commmand_commands(self):
        yield self.do_test_command('commands')
        self.assertEqual(self.sent[0][0], self.CHANNEL['id'])

    @defer.inlineCallbacks
    def test_commmand_commands_botfather(self):
        yield self.do_test_command('commands', 'botfather')
        self.assertEqual(self.sent[0][0], self.CHANNEL['id'])
        self.assertRegex(self.sent[0][1], r"^\w+ - \S+")

    @defer.inlineCallbacks
    def test_command_getid_private(self):
        yield self.do_test_command('getid', contact=self.contact1)
        self.assertEqual(len(self.sent), 1)
        self.assertIn(str(self.USER['id']), self.sent[0][1])

    @defer.inlineCallbacks
    def test_command_getid_group(self):
        yield self.do_test_command('getid')
        self.assertIn(str(self.USER['id']), self.sent[0][1])
        self.assertIn(str(self.CHANNEL['id']), self.sent[1][1])

    def assertButton(self, data, pos=None, sent=0):
        keyboard = self.sent[sent][2]['reply_markup']['inline_keyboard']
        if pos is not None:
            r, c = pos
            self.assertEqual(keyboard[r][c]['callback_data'], data)
        else:
            dataset = [b['callback_data'] for row in keyboard for b in row]
            self.assertIn(data, dataset)

    @defer.inlineCallbacks
    def test_command_list(self):
        yield self.do_test_command('list')
        self.assertButton('/list builders')
        self.assertButton('/list workers')
        self.assertButton('/list changes')

    @defer.inlineCallbacks
    def test_command_list_builders(self):
        yield self.do_test_command('list', 'all builders')
        self.assertEqual(len(self.sent), 1)
        for builder in self.BUILDER_NAMES:
            self.assertIn(f'`{builder}` ❌', self.sent[0][1])

    @defer.inlineCallbacks
    def test_command_list_workers(self):
        workers = ['worker1', 'worker2']
        for i, worker in enumerate(workers):
            yield self.master.db.workers.db.insert_test_data([fakedb.Worker(id=i, name=worker)])
        yield self.do_test_command('list', args='all workers')
        self.assertEqual(len(self.sent), 1)
        for worker in workers:
            self.assertIn(f'`{worker}` ❌', self.sent[0][1])

    @defer.inlineCallbacks
    def test_command_list_workers_online(self):
        self.setup_multi_builders()
        # Also set the connectedness:
        yield self.master.db.insert_test_data([
            fakedb.ConnectedWorker(id=113, masterid=13, workerid=1)
        ])
        yield self.do_test_command('list', args='all workers')
        self.assertEqual(len(self.sent), 1)
        self.assertNotIn('`linux1` ⚠️', self.sent[0][1])
        self.assertIn('`linux2` ⚠️', self.sent[0][1])

    @defer.inlineCallbacks
    def test_command_list_changes(self):
        yield self.master.db.workers.db.insert_test_data([
            fakedb.SourceStamp(id=14),
            fakedb.Change(changeid=99, sourcestampid=14),
        ])
        yield self.do_test_command('list', args='2 changes')
        self.assertEqual(len(self.sent), 2)

    @defer.inlineCallbacks
    def test_command_list_changes_long(self):
        yield self.master.db.workers.db.insert_test_data(
            [fakedb.SourceStamp(id=i) for i in range(1, 200)]
            + [fakedb.Change(changeid=i, sourcestampid=i) for i in range(1, 200)]
        )
        yield self.do_test_command('list', args='all changes')
        self.assertIn('reply_markup', self.sent[1][2])

    @defer.inlineCallbacks
    def test_command_watch(self):
        self.setupSomeBuilds()
        yield self.do_test_command('watch')
        self.assertButton('/watch builder1')

    @defer.inlineCallbacks
    def test_command_watch_no_builds(self):
        yield self.do_test_command('watch')

    @defer.inlineCallbacks
    def test_command_stop_no_args(self):
        self.setupSomeBuilds()
        yield self.do_test_command('stop')
        self.assertButton('/stop build builder1')

    @defer.inlineCallbacks
    def test_command_stop_ask_reason(self):
        self.patch_send()
        self.setupSomeBuilds()
        yield self.do_test_command('stop', 'build builder1')
        self.assertIn("give me the reason", self.sent[0][1])
        self.assertEqual(self.contact.template, "/stop build builder1 {}")

    def test_ask_reply_group(self):
        self.patch_send()
        self.contact.ask_for_reply("test")
        self.assertEqual(self.sent[0][1], "Ok @harrypotter, now test...")

    def test_ask_reply_group_no_username(self):
        self.patch_send()
        self.contact.user_info = self.USER.copy()
        del self.contact.user_info['username']
        self.contact.ask_for_reply("test")
        self.assertEqual(self.sent[0][1], "Ok, now reply to this message and test...")

    def test_ask_reply_group_no_username_no_greeting(self):
        self.patch_send()
        self.contact.user_info = self.USER.copy()
        del self.contact.user_info['username']
        self.contact.ask_for_reply("test", None)
        self.assertEqual(self.sent[0][1], "Reply to this message and test...")

    def test_ask_reply_private_no_greeting(self):
        self.patch_send()
        self.contact1.ask_for_reply("test", None)
        self.assertEqual(self.sent[0][1], "Test...")

    @defer.inlineCallbacks
    def test_command_notify_no_args(self):
        self.patch_send()
        self.contact.channel.notify_events = {'success', 'failure'}
        yield self.do_test_command('notify')
        self.assertButton('/notify on-quiet finished')
        self.assertButton('/notify off-quiet success')
        self.assertButton('/notify list')

    @defer.inlineCallbacks
    def test_command_notify_list_with_query(self):
        self.patch_send()

        def delete_message(chat, msg):
            delete_message.msg = msg

        delete_message.msg = None
        self.bot.delete_message = delete_message

        yield self.do_test_command('notify', 'list', tquery={'message': {'message_id': 2345}})
        self.assertEqual(delete_message.msg, 2345)

    @defer.inlineCallbacks
    def test_command_notify_toggle(self):
        self.patch_send()

        def edit_keyboard(chat, msg, keyboard):
            self.sent.append((chat, None, {'reply_markup': {'inline_keyboard': keyboard}}))

        self.bot.edit_keyboard = edit_keyboard

        self.contact.channel.notify_events = {'success', 'failure'}
        yield self.do_test_command(
            'notify', 'on-quiet finished', tquery={'message': {'message_id': 2345}}
        )
        self.assertIn('finished', self.contact.channel.notify_events)
        self.assertButton('/notify off-quiet finished')

    @defer.inlineCallbacks
    def test_command_shutdown(self):
        yield self.do_test_command('shutdown')
        self.assertButton('/shutdown start')
        self.assertButton('/shutdown now')

    @defer.inlineCallbacks
    def test_command_shutdown_shutting_down(self):
        yield self.do_test_command('shutdown', shuttingDown=True)
        self.assertButton('/shutdown stop')
        self.assertButton('/shutdown now')

    def allSchedulers(self):
        return self.schedulers

    @defer.inlineCallbacks
    def make_forcescheduler(self, two=False):
        scheduler = forcesched.ForceScheduler(
            name='force1',
            builderNames=['builder1', 'builder2'],
            codebases=[
                forcesched.CodebaseParameter(
                    '',
                    branch=forcesched.StringParameter(name='branch', default="master"),
                    repository=forcesched.FixedParameter(
                        name="repository", default="repository.git"
                    ),
                ),
                forcesched.CodebaseParameter(
                    'second',
                    branch=forcesched.StringParameter(name='branch', default="master"),
                    repository=forcesched.FixedParameter(
                        name="repository", default="repository2.git"
                    ),
                ),
            ],
            reason=forcesched.StringParameter(name='reason', required=True),
        )
        self.schedulers = [scheduler]
        if two:
            scheduler2 = forcesched.ForceScheduler(name='force2', builderNames=['builder2'])
            self.schedulers.append(scheduler2)
        self.bot.master.allSchedulers = self.allSchedulers
        for sched in self.schedulers:
            yield sched.setServiceParent(self.master)

    @defer.inlineCallbacks
    def test_command_force_no_schedulers(self):
        yield self.do_test_command('force', exp_UsageError=True)

    @defer.inlineCallbacks
    def test_command_force_noargs_multiple_schedulers(self):
        yield self.make_forcescheduler(two=True)
        yield self.do_test_command('force')
        self.assertButton('/force force1')
        self.assertButton('/force force2')

    @defer.inlineCallbacks
    def test_command_force_noargs(self):
        yield self.make_forcescheduler()
        yield self.do_test_command('force')
        self.assertButton('/force force1 config builder1')
        self.assertButton('/force force1 config builder2')

    @defer.inlineCallbacks
    def test_command_force_only_scheduler(self):
        yield self.make_forcescheduler()
        yield self.do_test_command('force', 'force1')
        self.assertButton('/force force1 config builder1')
        self.assertButton('/force force1 config builder2')

    @defer.inlineCallbacks
    def test_command_force_bad_scheduler(self):
        yield self.make_forcescheduler(two=True)
        yield self.do_test_command('force', 'force3', exp_UsageError=True)

    @defer.inlineCallbacks
    def test_command_force_bad_builder(self):
        yield self.make_forcescheduler()
        yield self.do_test_command('force', 'force1 config builder0', exp_UsageError=True)

    @defer.inlineCallbacks
    def test_command_force_bad_command(self):
        yield self.make_forcescheduler()
        yield self.do_test_command('force', 'force1 bad builder1', exp_UsageError=True)

    @defer.inlineCallbacks
    def test_command_force_only_bad_command(self):
        yield self.make_forcescheduler()
        yield self.do_test_command('force', 'bad builder1', exp_UsageError=True)

    @defer.inlineCallbacks
    def test_command_force_config(self):
        yield self.make_forcescheduler()
        yield self.do_test_command('force', 'force1 config builder1')
        self.assertButton('/force force1 ask reason builder1 ')
        self.assertButton('/force force1 ask branch builder1 ')
        self.assertButton('/force force1 ask project builder1 ')
        self.assertButton('/force force1 ask revision builder1 ')
        self.assertButton('/force force1 ask second_branch builder1 ')
        self.assertButton('/force force1 ask second_project builder1 ')
        self.assertButton('/force force1 ask second_revision builder1 ')

    @defer.inlineCallbacks
    def test_command_force_config_more(self):
        yield self.make_forcescheduler()
        yield self.do_test_command('force', 'force1 config builder1 branch=master')
        self.assertButton('/force force1 ask reason builder1 branch=master')

    @defer.inlineCallbacks
    def test_command_force_config_nothing_missing(self):
        yield self.make_forcescheduler()
        yield self.do_test_command('force', 'force1 config builder1 reason=Ok')
        self.assertButton('/force force1 build builder1 reason=Ok')

    @defer.inlineCallbacks
    def test_command_force_ask(self):
        yield self.make_forcescheduler()
        yield self.do_test_command('force', 'force1 ask reason builder1 branch=master')
        self.assertEqual(
            self.contact.template, '/force force1 config builder1 branch=master reason={}'
        )

    @defer.inlineCallbacks
    def test_command_force_build_missing(self):
        yield self.make_forcescheduler()
        yield self.do_test_command('force', 'force1 build builder1')
        self.assertButton('/force force1 ask reason builder1 ')

    @defer.inlineCallbacks
    def test_command_force_build(self):
        yield self.make_forcescheduler()

        force_args = {}

        def force(**kwargs):
            force_args.update(kwargs)

        self.schedulers[0].force = force

        yield self.do_test_command('force', 'force1 build builder1 reason=Good')
        self.assertEqual(self.sent[0][1], "Force build successfully requested.")

        expected = {
            'builderid': 23,
            'owner': "Harry Potter (@harrypotter) on 'Hogwards'",
            'reason': 'Good',
            'repository': 'repository.git',  # fixed param
            'second_repository': 'repository2.git',  # fixed param
        }
        self.assertEqual(force_args, expected)


class TestPollingBot(telegram.TelegramPollingBot):
    def __init__(self, updates, *args, **kwargs):
        self.__updates = updates
        super().__init__(*args, **kwargs)

    def process_update(self, update):
        self.__updates -= 1
        if not self.__updates:
            self._polling_continue = False
        return super().process_update(update)


class TestTelegramService(TestReactorMixin, unittest.TestCase):
    USER = TestTelegramContact.USER
    CHANNEL = TestTelegramContact.CHANNEL
    PRIVATE = TestTelegramContact.PRIVATE

    URL = 'https://api.telegram.org/bot12345:secret'

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.patch(reactor, 'callLater', self.reactor.callLater)
        self.master = yield fakemaster.make_master(self, wantData=True, wantDb=True, wantMq=True)
        self.http = None

    @defer.inlineCallbacks
    def setup_http_service(self):
        self.http = yield fakehttpclientservice.HTTPClientService.getService(
            self.master, self, self.URL
        )

    def setup_http_session(self):
        return httpclientservice.HTTPSession(self.master.httpservice, self.URL)

    @defer.inlineCallbacks
    def makeBot(self, chat_ids=None, authz=None, *args, **kwargs):
        if chat_ids is None:
            chat_ids = []
        if self.http is None:
            yield self.setup_http_service()
        www = get_plugins('www', None, load_now=True)
        if 'base' not in www:
            raise SkipTest('telegram tests need buildbot-www installed')
        return telegram.TelegramWebhookBot(
            '12345:secret', self.setup_http_session(), chat_ids, authz, *args, **kwargs
        )

    @defer.inlineCallbacks
    def test_getContact(self):
        bot = yield self.makeBot()
        c1 = bot.getContact(self.USER, self.PRIVATE)
        c2 = bot.getContact(self.USER, self.CHANNEL)
        c1b = bot.getContact(self.USER, self.PRIVATE)
        self.assertIs(c1, c1b)
        self.assertIsInstance(c2, words.Contact)
        self.assertIn((-12345678, 123456789), bot.contacts)
        self.assertEqual({123456789, -12345678}, set(bot.channels.keys()))

    @defer.inlineCallbacks
    def test_getContact_update(self):
        try:
            bot = yield self.makeBot()
            contact = bot.getContact(self.USER, self.CHANNEL)
            updated_user = self.USER.copy()
            updated_user['username'] = "dirtyharry"
            self.assertEqual(contact.user_info['username'], "harrypotter")
            bot.getContact(updated_user, self.CHANNEL)
            self.assertEqual(contact.user_info['username'], "dirtyharry")
        finally:
            self.USER['username'] = "harrypotter"

    @defer.inlineCallbacks
    def test_getContact_invalid(self):
        bot = yield self.makeBot()
        bot.authz = {'': None}

        u = bot.getContact(user=self.USER, channel=self.CHANNEL)
        self.assertNotIn((-12345678, 123456789), bot.contacts)
        self.assertNotIn(-12345678, bot.channels)

        if platform.python_implementation() != 'PyPy':
            self.assertEqual(sys.getrefcount(u), 2)  # local, sys
            c = u.channel
            self.assertEqual(sys.getrefcount(c), 3)  # local, contact, sys
            del u
            self.assertEqual(sys.getrefcount(c), 2)  # local, sys

    @defer.inlineCallbacks
    def test_getContact_valid(self):
        bot = yield self.makeBot()
        bot.authz = {'': None, 'command': 123456789}

        bot.getContact(user=self.USER, channel=self.CHANNEL)
        self.assertIn((-12345678, 123456789), bot.contacts)

    @defer.inlineCallbacks
    def test_set_webhook(self):
        bot = yield self.makeBot()
        self.http.expect("post", "/setWebhook", json={'url': 'our.webhook'}, content_json={'ok': 1})
        yield bot.set_webhook('our.webhook')

    @defer.inlineCallbacks
    def test_set_webhook_cert(self):
        bot = yield self.makeBot()
        self.http.expect(
            "post",
            "/setWebhook",
            data={'url': 'our.webhook'},
            files={'certificate': b"this is certificate"},
            content_json={'ok': 1},
        )
        yield bot.set_webhook('our.webhook', "this is certificate")

    @defer.inlineCallbacks
    def test_send_message(self):
        bot = yield self.makeBot()
        self.http.expect(
            "post",
            "/sendMessage",
            json={'chat_id': 1234, 'text': 'Hello', 'parse_mode': 'Markdown'},
            content_json={'ok': 1, 'result': {'message_id': 9876}},
        )
        m = yield bot.send_message(1234, 'Hello')
        self.assertEqual(m['message_id'], 9876)

    @defer.inlineCallbacks
    def test_send_message_long(self):
        bot = yield self.makeBot()

        text1 = '\n'.join(f"{i + 1:039}" for i in range(102))
        text2 = '\n'.join(f"{i + 1:039}" for i in range(102, 204))
        text3 = '\n'.join(f"{i + 1:039}" for i in range(204, 250))

        self.http.expect(
            "post",
            "/sendMessage",
            json={
                'chat_id': 1234,
                'text': text1,
                'parse_mode': 'Markdown',
                'reply_to_message_id': 1000,
            },
            content_json={'ok': 1, 'result': {'message_id': 1001}},
        )
        self.http.expect(
            "post",
            "/sendMessage",
            json={'chat_id': 1234, 'text': text2, 'parse_mode': 'Markdown'},
            content_json={'ok': 1, 'result': {'message_id': 1002}},
        )
        self.http.expect(
            "post",
            "/sendMessage",
            json={
                'chat_id': 1234,
                'text': text3,
                'parse_mode': 'Markdown',
                'reply_markup': {'inline_keyboard': 'keyboard'},
            },
            content_json={'ok': 1, 'result': {'message_id': 1003}},
        )

        text = '\n'.join(f"{i + 1:039}" for i in range(250))
        m = yield bot.send_message(
            1234, text, reply_markup={'inline_keyboard': 'keyboard'}, reply_to_message_id=1000
        )
        self.assertEqual(m['message_id'], 1003)

    @defer.inlineCallbacks
    def test_edit_message(self):
        bot = yield self.makeBot()
        self.http.expect(
            "post",
            "/editMessageText",
            json={'chat_id': 1234, 'message_id': 9876, 'text': 'Hello', 'parse_mode': 'Markdown'},
            content_json={'ok': 1, 'result': {'message_id': 9876}},
        )
        m = yield bot.edit_message(1234, 9876, 'Hello')
        self.assertEqual(m['message_id'], 9876)

    @defer.inlineCallbacks
    def test_delete_message(self):
        bot = yield self.makeBot()
        self.http.expect(
            "post",
            "/deleteMessage",
            json={'chat_id': 1234, 'message_id': 9876},
            content_json={'ok': 1},
        )
        yield bot.delete_message(1234, 9876)

    @defer.inlineCallbacks
    def test_send_sticker(self):
        bot = yield self.makeBot()
        self.http.expect(
            "post",
            "/sendSticker",
            json={'chat_id': 1234, 'sticker': 'xxxxx'},
            content_json={'ok': 1, 'result': {'message_id': 9876}},
        )
        m = yield bot.send_sticker(1234, 'xxxxx')
        self.assertEqual(m['message_id'], 9876)

    @defer.inlineCallbacks
    def test_set_nickname(self):
        bot = yield self.makeBot()
        self.assertIsNone(bot.nickname)
        self.http.expect(
            "post", "/getMe", content_json={'ok': 1, 'result': {'username': 'testbot'}}
        )
        yield bot.set_nickname()
        self.assertEqual(bot.nickname, 'testbot')

    def prepare_request(self, **kwargs):
        payload = {"update_id": 12345}
        payload.update(kwargs)
        content = unicode2bytes(json.dumps(payload))
        request = FakeRequest(content=content)
        request.uri = b"/bot12345:secret"
        request.method = b"POST"
        request.received_headers[b'Content-Type'] = b"application/json"
        return request

    def request_message(self, text):
        return self.prepare_request(
            message={
                "message_id": 123,
                "from": self.USER,
                "chat": self.CHANNEL,
                "date": 1566688888,
                "text": text,
            }
        )

    def request_query(self, data):
        return self.prepare_request(
            callback_query={
                "id": 123456,
                "from": self.USER,
                "data": data,
                "message": {
                    "message_id": 12345,
                    "from": self.USER,
                    "chat": self.CHANNEL,
                    "date": 1566688888,
                },
            }
        )

    @defer.inlineCallbacks
    def test_get_update(self):
        bot = yield self.makeBot()
        request = self.request_message("test")
        update = bot.get_update(request)
        self.assertEqual(update['message']['from'], self.USER)
        self.assertEqual(update['message']['chat'], self.CHANNEL)

    @defer.inlineCallbacks
    def test_get_update_bad_content_type(self):
        bot = yield self.makeBot()
        request = self.request_message("test")
        request.received_headers[b'Content-Type'] = b"application/data"
        with self.assertRaises(ValueError):
            bot.get_update(request)

    @defer.inlineCallbacks
    def test_render_POST(self):
        # This actually also tests process_incoming
        bot = yield self.makeBot()
        bot.contactClass = FakeContact
        request = self.request_message("test")
        bot.webhook.render_POST(request)
        contact = bot.getContact(self.USER, self.CHANNEL)
        self.assertEqual(contact.messages, ["test"])

    @defer.inlineCallbacks
    def test_parse_query_cached(self):
        bot = yield self.makeBot()
        bot.contactClass = FakeContact
        bot.query_cache.update({100: "good"})
        self.http.expect(
            "post",
            "/answerCallbackQuery",
            json={'callback_query_id': 123456},
            content_json={'ok': 1},
        )
        request = self.request_query("100")
        bot.process_webhook(request)
        self.assertEqual(bot.getContact(self.USER, self.CHANNEL).messages, ["good"])

    @defer.inlineCallbacks
    def test_parse_query_cached_dict(self):
        bot = yield self.makeBot()
        bot.contactClass = FakeContact
        bot.query_cache = {100: {'command': "good", 'notify': "hello"}}
        self.http.expect(
            "post",
            "/answerCallbackQuery",
            json={'callback_query_id': 123456, 'text': "hello"},
            content_json={'ok': 1},
        )
        request = self.request_query("100")
        bot.process_webhook(request)
        self.assertEqual(bot.getContact(self.USER, self.CHANNEL).messages, ["good"])

    @defer.inlineCallbacks
    def test_parse_query_explicit(self):
        bot = yield self.makeBot()
        bot.contactClass = FakeContact
        bot.query_cache = {100: "bad"}
        self.http.expect(
            "post",
            "/answerCallbackQuery",
            json={'callback_query_id': 123456},
            content_json={'ok': 1},
        )
        request = self.request_query("good")
        bot.process_webhook(request)
        self.assertEqual(bot.getContact(self.USER, self.CHANNEL).messages, ["good"])

    @defer.inlineCallbacks
    def test_parse_query_bad(self):
        bot = yield self.makeBot()
        bot.contactClass = FakeContact
        bot.query_cache.update({100: "bad"})
        self.http.expect(
            "post",
            "/editMessageReplyMarkup",
            json={'chat_id': -12345678, 'message_id': 12345},
            content_json={'ok': 1},
        )
        self.http.expect(
            "post",
            "/answerCallbackQuery",
            json={'callback_query_id': 123456, 'text': "Sorry, button is no longer valid!"},
            content_json={'ok': 1},
        )
        request = self.request_query("101")
        bot.process_webhook(request)

    def makePollingBot(self, updates, chat_ids=None, authz=None, *args, **kwargs):
        if chat_ids is None:
            chat_ids = []

        return TestPollingBot(
            updates, '12345:secret', self.setup_http_session(), chat_ids, authz, *args, **kwargs
        )

    @defer.inlineCallbacks
    def test_polling(self):
        yield self.setup_http_service()
        bot = self.makePollingBot(2)
        bot._polling_continue = True
        self.http.expect("post", "/deleteWebhook", content_json={"ok": 1})
        self.http.expect(
            "post",
            "/getUpdates",
            json={'timeout': bot.poll_timeout},
            content_json={
                'ok': 1,
                'result': [
                    {
                        "update_id": 10000,
                        "message": {
                            "message_id": 123,
                            "from": self.USER,
                            "chat": self.CHANNEL,
                            "date": 1566688888,
                            "text": "ignore",
                        },
                    }
                ],
            },
        )
        self.http.expect(
            "post",
            "/getUpdates",
            json={'timeout': bot.poll_timeout, "offset": 10001},
            content_json={
                'ok': 1,
                'result': [
                    {
                        "update_id": 10001,
                        "message": {
                            "message_id": 124,
                            "from": self.USER,
                            "chat": self.CHANNEL,
                            "date": 1566688889,
                            "text": "/nay",
                        },
                    }
                ],
            },
        )
        self.http.expect(
            "post",
            "/sendMessage",
            json={'chat_id': -12345678, 'text': 'Never mind, Harry...', 'parse_mode': 'Markdown'},
            content_json={'ok': 1, 'result': {'message_id': 125}},
        )
        yield bot.do_polling()

    @defer.inlineCallbacks
    def test_format_build_status(self):
        bot = yield self.makeBot()
        build = {'results': SUCCESS}
        self.assertEqual(bot.format_build_status(build), "completed successfully ✅")

    @defer.inlineCallbacks
    def test_format_build_status_short(self):
        bot = yield self.makeBot()
        build = {'results': WARNINGS}
        self.assertEqual(bot.format_build_status(build, short=True), " ⚠️")

    class HttpServiceWithErrors(fakehttpclientservice.HTTPClientService):
        def __init__(self, skip, errs, *args, **kwargs):
            self.__skip = skip
            self.__errs = errs
            self.succeeded = False
            super().__init__(*args, **kwargs)

        def _do_request(self, session, method, ep, **kwargs):
            if method == 'post':
                if self.__skip:
                    self.__skip -= 1
                else:
                    if self.__errs:
                        self.__errs -= 1
                        raise RuntimeError(f"{self.__errs + 1}")
                    self.succeeded = True
            return super()._do_request(session, method, ep, **kwargs)

    # returns a Deferred
    @defer.inlineCallbacks
    def setup_http_service_with_errors(self, skip, errs):
        url = 'https://api.telegram.org/bot12345:secret'
        self.http = yield self.HttpServiceWithErrors.getService(self.master, self, skip, errs, url)

    @defer.inlineCallbacks
    def test_post_not_ok(self):
        bot = yield self.makeBot()
        self.http.expect("post", "/post", content_json={'ok': 0})

        def log(msg):
            logs.append(msg)

        logs = []
        bot.log = log

        yield bot.post("/post")
        self.assertIn("ERROR", logs[0])

    @defer.inlineCallbacks
    def test_post_need_repeat(self):
        yield self.setup_http_service_with_errors(0, 2)
        bot = yield self.makeBot()
        self.http.expect("post", "/post", content_json={'ok': 1})

        def log(msg):
            logs.append(msg)

        logs = []
        bot.log = log

        bot.post("/post")
        self.assertIn("ERROR", logs[0])

        self.reactor.pump(3 * [30.0])

        self.assertTrue(self.http.succeeded)

    @defer.inlineCallbacks
    def test_polling_need_repeat(self):
        yield self.setup_http_service_with_errors(1, 2)
        bot = self.makePollingBot(1)
        bot.reactor = self.reactor
        bot._polling_continue = True
        self.http.expect("post", "/deleteWebhook", content_json={"ok": 1})
        self.http.expect(
            "post",
            "/getUpdates",
            json={'timeout': bot.poll_timeout},
            content_json={
                'ok': 1,
                'result': [
                    {
                        "update_id": 10000,
                        "message": {
                            "message_id": 123,
                            "from": self.USER,
                            "chat": self.CHANNEL,
                            "date": 1566688888,
                            "text": "ignore",
                        },
                    }
                ],
            },
        )

        def log(msg):
            logs.append(msg)

        logs = []
        bot.log = log

        bot.do_polling()
        self.assertIn("ERROR", logs[0])

        self.reactor.pump(3 * [30.0])

        self.assertTrue(self.http.succeeded)



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


from buildbot.test.fakedb.row import Row


class Log(Row):
    table = "logs"

    id_column = 'id'

    def __init__(
        self, id=None, name='log29', slug=None, stepid=None, complete=0, num_lines=0, type='s'
    ):
        if slug is None:
            slug = name
        super().__init__(
            id=id,
            name=name,
            slug=slug,
            stepid=stepid,
            complete=complete,
            num_lines=num_lines,
            type=type,
        )


class LogChunk(Row):
    table = "logchunks"

    # 'content' column is sa.LargeBinary, it's bytestring.
    binary_columns = ('content',)

    def __init__(self, logid=None, first_line=0, last_line=0, content='', compressed=0):
        super().__init__(
            logid=logid,
            first_line=first_line,
            last_line=last_line,
            content=content,
            compressed=compressed,
        )



# This file is part of Buildbot.  Buildbot is free software: you can
# redistribute it and/or modify it under the terms of the GNU General Public
# License as published by the Free Software Foundation, version 2.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc., 51
# Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Copyright Buildbot Team Members


import os

from twisted.cred import credentials
from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.endpoints import clientFromString
from twisted.python import log
from twisted.python import util
from twisted.spread import pb
from twisted.trial import unittest

import buildbot
from buildbot import config
from buildbot import worker
from buildbot.process import botmaster
from buildbot.process import builder
from buildbot.process import factory
from buildbot.test.fake import fakemaster
from buildbot.test.reactor import TestReactorMixin
from buildbot.util.eventual import eventually
from buildbot.worker import manager as workermanager
from buildbot.worker.protocols.manager.pb import PBManager

PKI_DIR = util.sibpath(__file__, 'pki')


class FakeWorkerForBuilder(pb.Referenceable):
    """
    Fake worker-side WorkerForBuilder object
    """


class FakeWorkerWorker(pb.Referenceable):
    """
    Fake worker-side Worker object

    @ivar master_persp: remote perspective on the master
    """

    def __init__(self, callWhenBuilderListSet):
        self.callWhenBuilderListSet = callWhenBuilderListSet
        self.master_persp = None
        self._detach_deferreds = []
        self._detached = False

    def waitForDetach(self):
        if self._detached:
            return defer.succeed(None)
        d = defer.Deferred()
        self._detach_deferreds.append(d)
        return d

    def setMasterPerspective(self, persp):
        self.master_persp = persp
        # clear out master_persp on disconnect

        def clear_persp():
            self.master_persp = None

        persp.broker.notifyOnDisconnect(clear_persp)

        def fire_deferreds():
            self._detached = True
            deferreds = self._detach_deferreds
            self._detach_deferreds = None
            for d in deferreds:
                d.callback(None)

        persp.broker.notifyOnDisconnect(fire_deferreds)

    def remote_print(self, message):
        log.msg(f"WORKER-SIDE: remote_print({message!r})")

    def remote_getWorkerInfo(self):
        return {
            'info': 'here',
            'worker_commands': {
                'x': 1,
            },
            'numcpus': 1,
            'none': None,
            'os_release': b'\xe3\x83\x86\xe3\x82\xb9\xe3\x83\x88'.decode(),
            b'\xe3\x83\xaa\xe3\x83\xaa\xe3\x83\xbc\xe3\x82\xb9\xe3'
            b'\x83\x86\xe3\x82\xb9\xe3\x83\x88'.decode(): b'\xe3\x83\x86\xe3\x82\xb9\xe3\x83\x88'.decode(),
        }

    def remote_getVersion(self):
        return buildbot.version

    def remote_getCommands(self):
        return {'x': 1}

    def remote_setBuilderList(self, builder_info):
        builder_names = [n for n, dir in builder_info]
        slbuilders = [FakeWorkerForBuilder() for n in builder_names]
        eventually(self.callWhenBuilderListSet)
        return dict(zip(builder_names, slbuilders))


class FakeBuilder(builder.Builder):
    def attached(self, worker, commands):
        return defer.succeed(None)

    def detached(self, worker):
        pass

    def getOldestRequestTime(self):
        return 0

    def maybeStartBuild(self):
        return defer.succeed(None)


class MyWorker(worker.Worker):
    def attached(self, conn):
        self.detach_d = defer.Deferred()
        return super().attached(conn)

    def detached(self):
        super().detached()
        d = self.detach_d
        self.detach_d = None
        d.callback(None)


class TestWorkerComm(TestReactorMixin, unittest.TestCase):
    """
    Test handling of connections from workers as integrated with
     - Twisted Spread
     - real TCP connections.
     - PBManager

    @ivar master: fake build master
    @ivar pbamanger: L{PBManager} instance
    @ivar botmaster: L{BotMaster} instance
    @ivar worker: master-side L{Worker} instance
    @ivar workerworker: worker-side L{FakeWorkerWorker} instance
    @ivar port: TCP port to connect to
    @ivar server_connection_string: description string for the server endpoint
    @ivar client_connection_string_tpl: description string template for the client
                                endpoint (expects to passed 'port')
    @ivar endpoint: endpoint controlling the outbound connection
                    from worker to master
    """

    @defer.inlineCallbacks
    def setUp(self):
        self.setup_test_reactor()
        self.master = yield fakemaster.make_master(self, wantMq=True, wantData=True, wantDb=True)

        # set the worker port to a loopback address with unspecified
        # port
        self.pbmanager = self.master.pbmanager = PBManager()
        yield self.pbmanager.setServiceParent(self.master)

        # remove the fakeServiceParent from fake service hierarchy, and replace
        # by a real one
        yield self.master.workers.disownServiceParent()
        self.workers = self.master.workers = workermanager.WorkerManager(self.master)
        yield self.workers.setServiceParent(self.master)

        self.botmaster = botmaster.BotMaster()
        yield self.botmaster.setServiceParent(self.master)

        self.master.botmaster = self.botmaster
        self.master.data.updates.workerConfigured = lambda *a, **k: None
        yield self.master.startService()

        self.buildworker = None
        self.port = None
        self.workerworker = None
        self.endpoint = None
        self.broker = None
        self._detach_deferreds = []

        # patch in our FakeBuilder for the regular Builder class
        self.patch(botmaster, 'Builder', FakeBuilder)

        self.server_connection_string = "tcp:0:interface=127.0.0.1"
        self.client_connection_string_tpl = "tcp:host=127.0.0.1:port={port}"

        @defer.inlineCallbacks
        def cleanup():
            if self.broker:
                del self.broker
            if self.endpoint:
                del self.endpoint
            deferreds = [
                *self._detach_deferreds,
                self.pbmanager.stopService(),
                self.botmaster.stopService(),
                self.workers.stopService(),
            ]

            # if the worker is still attached, wait for it to detach, too
            if self.buildworker and self.buildworker.detach_d:
                deferreds.append(self.buildworker.detach_d)

            yield defer.gatherResults(deferreds, consumeErrors=True)

        self.addCleanup(cleanup)

    @defer.inlineCallbacks
    def addWorker(self, **kwargs):
        """
        Create a master-side worker instance and add it to the BotMaster

        @param **kwargs: arguments to pass to the L{Worker} constructor.
        """
        self.buildworker = MyWorker("testworker", "pw", **kwargs)

        # reconfig the master to get it set up
        new_config = self.master.config
        new_config.protocols = {"pb": {"port": self.server_connection_string}}
        new_config.workers = [self.buildworker]
        new_config.builders = [
            config.BuilderConfig(
                name='bldr', workername='testworker', factory=factory.BuildFactory()
            )
        ]

        yield self.botmaster.reconfigServiceWithBuildbotConfig(new_config)
        yield self.workers.reconfigServiceWithBuildbotConfig(new_config)

        # as part of the reconfig, the worker registered with the pbmanager, so
        # get the port it was assigned
        self.port = self.buildworker.registration.getPBPort()

    def connectWorker(self, waitForBuilderList=True):
        """
        Connect a worker the master via PB

        @param waitForBuilderList: don't return until the setBuilderList has
        been called
        @returns: L{FakeWorkerWorker} and a Deferred that will fire when it
        is detached; via deferred
        """
        factory = pb.PBClientFactory()
        creds = credentials.UsernamePassword(b"testworker", b"pw")
        setBuilderList_d = defer.Deferred()
        workerworker = FakeWorkerWorker(lambda: setBuilderList_d.callback(None))

        login_d = factory.login(creds, workerworker)

        @login_d.addCallback
        def logged_in(persp):
            workerworker.setMasterPerspective(persp)

            # set up to hear when the worker side disconnects
            workerworker.detach_d = defer.Deferred()
            persp.broker.notifyOnDisconnect(lambda: workerworker.detach_d.callback(None))
            self._detach_deferreds.append(workerworker.detach_d)

            return workerworker

        self.endpoint = clientFromString(
            reactor, self.client_connection_string_tpl.format(port=self.port)
        )
        connected_d = self.endpoint.connect(factory)

        dlist = [connected_d, login_d]
        if waitForBuilderList:
            dlist.append(setBuilderList_d)

        d = defer.DeferredList(dlist, consumeErrors=True, fireOnOneErrback=True)
        d.addCallback(lambda _: workerworker)
        return d

    def workerSideDisconnect(self, worker):
        """Disconnect from the worker side"""
        worker.master_persp.broker.transport.loseConnection()

    @defer.inlineCallbacks
    def test_connect_disconnect(self):
        """Test a single worker connecting and disconnecting."""
        yield self.addWorker()

        # connect
        worker = yield self.connectWorker()

        # disconnect
        self.workerSideDisconnect(worker)

        # wait for the resulting detach
        yield worker.waitForDetach()

    @defer.inlineCallbacks
    def test_tls_connect_disconnect(self):
        """Test with TLS or SSL endpoint.

        According to the deprecation note for the SSL client endpoint,
        the TLS endpoint is supported from Twistd 16.0.

        TODO add certificate verification (also will require some conditionals
        on various versions, including PyOpenSSL, service_identity. The CA used
        to generate the testing cert is in ``PKI_DIR/ca``
        """

        def escape_colon(path):
            # on windows we can't have \ as it serves as the escape character for :
            return path.replace('\\', '/').replace(':', '\\:')

        self.server_connection_string = (
            "ssl:port=0:certKey={pub}:privateKey={priv}:" + "interface=127.0.0.1"
        ).format(
            pub=escape_colon(os.path.join(PKI_DIR, '127.0.0.1.crt')),
            priv=escape_colon(os.path.join(PKI_DIR, '127.0.0.1.key')),
        )
        self.client_connection_string_tpl = "ssl:host=127.0.0.1:port={port}"

        yield self.addWorker()

        # connect
        worker = yield self.connectWorker()

        # disconnect
        self.workerSideDisconnect(worker)

        # wait for the resulting detach
        yield worker.waitForDetach()

    @defer.inlineCallbacks
    def test_worker_info(self):
        yield self.addWorker()
        worker = yield self.connectWorker()
        props = self.buildworker.info
        # check worker info passing
        self.assertEqual(props.getProperty("info"), "here")
        # check worker info passing with UTF-8
        self.assertEqual(
            props.getProperty("os_release"), b'\xe3\x83\x86\xe3\x82\xb9\xe3\x83\x88'.decode()
        )
        self.assertEqual(
            props.getProperty(
                b'\xe3\x83\xaa\xe3\x83\xaa\xe3\x83\xbc\xe3\x82'
                b'\xb9\xe3\x83\x86\xe3\x82\xb9\xe3\x83\x88'.decode()
            ),
            b'\xe3\x83\x86\xe3\x82\xb9\xe3\x83\x88'.decode(),
        )
        self.assertEqual(props.getProperty("none"), None)
        self.assertEqual(props.getProperty("numcpus"), 1)

        self.workerSideDisconnect(worker)
        yield worker.waitForDetach()

    @defer.inlineCallbacks
    def _test_duplicate_worker(self):
        yield self.addWorker()

        # connect first worker
        worker1 = yield self.connectWorker()

        # connect second worker; this should fail
        try:
            yield self.connectWorker(waitForBuilderList=False)
            connect_failed = False
        except Exception:
            connect_failed = True
        self.assertTrue(connect_failed)

        # disconnect both and wait for that to percolate
        self.workerSideDisconnect(worker1)

        yield worker1.waitForDetach()

        # flush the exception logged for this on the master
        self.assertEqual(len(self.flushLoggedErrors(RuntimeError)), 1)

    @defer.inlineCallbacks
    def _test_duplicate_worker_old_dead(self):
        yield self.addWorker()

        # connect first worker
        worker1 = yield self.connectWorker()

        # monkeypatch that worker to fail with PBConnectionLost when its
        # remote_print method is called
        def remote_print(message):
            worker1.master_persp.broker.transport.loseConnection()
            raise pb.PBConnectionLost("fake!")

        worker1.remote_print = remote_print

        # connect second worker; this should succeed, and the old worker
        # should be disconnected.
        worker2 = yield self.connectWorker()

        # disconnect both and wait for that to percolate
        self.workerSideDisconnect(worker2)

        yield worker1.waitForDetach()

        # flush the exception logged for this on the worker
        self.assertEqual(len(self.flushLoggedErrors(pb.PBConnectionLost)), 1)
